import gitlab
import urllib
import json
import base64
import re
import os

# private token or personal token authentication

class gitLabFile(object):
    
    @staticmethod
    def get_gitlab_file(uri, read_file=True):
        
        gitlab_repository = os.environ.get('OBSINFO_GITLAB_REPOSITORY')
        project_path = os.environ.get('OBSINFO_PROJECTPATH')
        
        if not gitlab_repository or not project_path:
            raise ValueError("One or several environment variables, OBSINFO_GITLAB_REPOSITORY or \
                      OBSINFO_PROJECTPATH, are missing")
        
        pattern = gitlab_repository + "/"
        uri = re.sub(pattern, "", uri)
        
        gl = gitlab.Gitlab("http://" + gitlab_repository)
        
        project = gl.projects.get(project_path)
        #uri = urllib.parse.urlsplit(str(uri)).path
        #if uri[0] == '/': # gitlab.projects.get doesn't like an absolute path'
        #    uri = uri[1:]

        try:
            f = project.files.get(file_path=uri, ref='v0.109')
            if not read_file:
                return True
        except gitlab.exceptions.GitlabOperationError:
            if read_file:
                print(f'Error reading remote (gitlab) file: {uri}')
                raise FileNotFoundError(uri)
                return None               
            else:
                return False
        # get the decoded content. Two decodes are necessary, from 8character string to bytes and 
        # from bytes to a python string
        bytecontent = base64.b64decode(f.content)
        ret = bytecontent.decode("utf-8")

        return ret
    
    @staticmethod
    def isRemote(file):
        return re.match('^http:', file) or re.match('^https:', file)

#obsinfo/_examples2/Information_Files/network/BBOBS.INSU-IPGP,network.yaml/raw?ref=v0.109')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functions to test the lcheapo functions
There are two types of testing functionalities.
a) If the file name includes "test--attributes" the output of the corresonding obsinfo test function
will be checked against data contained in this class.
b) If the fine name is "normal", it will simply run through to make sure there are no errors
Testing for filters uses data from obsinfo/tests/data/Information_Files/responses/_filters. 
Testing for all other classes uses data from obsinfo/_examples/Information_Files 
WARNING: many tests are crtically dependent on file hierarchy, including names. Do not change names
in tests or _examples hierarchy, or else change the names here.
Also, the following methods use four specific filen names:
test_all_stage_types()
test_senso()
test_preamplifier()
test_datalogger()
test_station()
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import os
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
import sys
from pathlib import Path, PurePath
import unittest
import inspect
import difflib
import re
import glob
from json.decoder import JSONDecodeError

from obspy.core.utcdatetime import UTCDateTime
# from pprint import pprint
# import xml.etree.ElementTree as ET
# from CompareXMLTree import XmlTree
from obsinfo.obsMetadata.obsmetadata import (ObsMetadata)
from obsinfo.instrumentation import (Instrumentation, InstrumentComponent,
                                     Datalogger, Preamplifier, Sensor,
                                     ResponseStages, Stage, Filter)
from obsinfo.instrumentation.instrumentation import (Location)
from obsinfo.network import (Station, Network)
from obsinfo.instrumentation.filter import (Filter, PolesZeros, FIR, Coefficients, ResponseList,
                     Analog, Digital, AD_Conversion)
from obsinfo.misc.printobs import  (PrintObs)
from obsinfo.misc.discoveryfiles import Datapath
import obsinfo

class JsonRefTest(unittest.TestCase):
    
    def assertTextFilesEqual(self, first, second, msg=None):
        with open(first) as f:
            str_a = f.read()
        with open(second) as f:
            str_b = f.read()

        if str_a != str_b:
            first_lines = str_a.splitlines(True)
            second_lines = str_b.splitlines(True)
            delta = difflib.unified_diff(
                first_lines, second_lines,
                fromfile=first, tofile=second)
            message = ''.join(delta)

            if msg:
                message += " : " + msg

            self.fail("Multi-line strings are unequal:\n" + message)

    def test_readJSONREF_json(self):
        """
        Test JSONref using a JSON file.
        """
        fname_A =  PurePath(self.testing_path).joinpath("jsonref_A.json")
        A = ObsMetadata.read_json_yaml_ref(fname_A, None)
        AB = ObsMetadata.read_json_yaml_ref(PurePath(self.testing_path).joinpath(
                                              "jsonref_AB.json"), None)
        self.assertTrue(A == AB)

    def test_readJSONREF_yaml(self):
        """
        Test JSONref using a YAML file.
        """
        A = ObsMetadata.read_json_yaml_ref(PurePath(self.testing_path).joinpath(
                                             "jsonref_A.yaml"), None)
        AB = ObsMetadata.read_json_yaml_ref(PurePath(self.testing_path).joinpath(
                                              "jsonref_AB.yaml"), None)
        self.assertTrue(A == AB)

    def test_validate_json(self):
        """
        Test validation on a YAML file.

        The test file as an $ref to a file that doesn't exist, a field that
        is not specified in the the schema, and lacks a field required in
        the schema
        """
        test_file = PurePath(self.testing_path).joinpath( 'json_testschema.json')
        test_schema = PurePath(self.testing_path).joinpath(
                                   'json_testschema.schema.json')
        # self.assertFalse(validate(test_file, schema_file=test_schema,
        #                           quiet=True))

        # Run the code
        cmd = f'obsinfo-validate -s {test_schema} {test_file} > temp'
        os.system(cmd)

        # Compare text files
        self.assertTextFilesEqual(
            'temp',
            PurePath(self.testing_path).joinpath( 'json_testschema.out.txt')
            )
        os.remove('temp')
          

class TestObsinfo(unittest.TestCase):
    """
    Test suite for obsinfo operations.
    """
    def setUp(self, verbose=True, test=True, print_output=False, level=None, ):
        """
        Set up default values and paths
        """   
        self.infofiles_path =  Datapath()
    
        self.level = level
        self.verbose = verbose
        self.test = test
        self.print_output = print_output
        
    def test_all_filters(self):
        """
        Test all information files in test/data/Information_Files/responses/_filters" subdirectory
        If you wish to test individual files, use test_filter(file) with file an absolute or 
        relative file name.
        """
        for dir in self.infofiles_path.datapath_list:
            files_in_validate_dir = Path(dir).joinpath(
                "*rs", # includes sensors, preamplifiers and dataloggers
                "responses",
                "filters/*.yaml")
            
            filelist = glob.glob(str(files_in_validate_dir))
            
            for file in filelist:
                self._test_filter(file)
            
            #if self.test:
            #   self._test_PZ_conditionals()
                    

    def _test_filter(self, info_file):
        """
        Test reading a filter file. All are actual examples except the info files called "test---attributes. 
        In this special cases there will also be a comparison against a dict of expected results
        """
        test_expected_result = { 
            'PolesZeros' : 
                {
                    "type": "PolesZeros",
                    "transfer_function_type": "LAPLACE (RADIANS/SECOND)",
                    "zeros": [(0.0+0.0j)],
                    "poles": [(0.546+0.191j), (4.00004e4+0.000j),],
                    "normalization_frequency" : 1.,
                    "normalization_factor" : 42833.458122775904,
                    "offset": 0,  
                },
            'FIR' : 
                {
                    "type": "FIR",
                    "symmetry": "ODD",
                    "coefficient_divisor": 8388608,
                    "coefficients": [-10944, 0, 103807, 0, -507903, 0, 2512192, 4194304,],
                    "offset": 7,  
                },
            'Coefficients' : 
                {
                    "type" : "Coefficients",
                    "transfer_function_type" : "DIGITAL", 
                    "numerator_coefficients" :   [1, 0.1, -0.3, 0.6],
                    "denominator_coefficients" : [-0.2, 0.8, 0.4, -0.3],
                    "offset": 0,
                },
            'ResponseList' : 
                {
                    "type" : "ResponseList",
                    "offset": 0,  
                    "response_list" : [[ 0.050, 0.56, 0.0], [ 0.075, 0.73, 0.0], [ 1, 1, 0.0], [10, 0.97, -179], [100, 0.96, 179], [1000, 0.96, 179], [5000, 0.82, 143], [7500, 0.69, 129]],
                },
            'AD_CONVERSION' : 
                {
                    "type" : "AD_CONVERSION",
                    "input_full_scale" : 5,
                    "output_full_scale" : 4294967292,
                    "transfer_function_type" : "DIGITAL",
                    "numerator_coefficients" :   [1.0],
                    "denominator_coefficients" : [],
                    "offset": 0,  
                },
            'ANALOG' : 
                {
                    "type" : "ANALOG",
                    "transfer_function_type" : "LAPLACE (RADIANS/SECOND)",
                    "zeros": [],
                    "poles": [],
                    "normalization_frequency" : 0.,
                    "normalization_factor" : 1.0,
                    "offset": 0,      
                },
            'DIGITAL' : 
                {
                    "type" : "DIGITAL",
                    "transfer_function_type" : "DIGITAL",
                    "numerator_coefficients" :   [1.0],
                    "denominator_coefficients" : [],
                    "offset": 0,                   
                },
             }
            
        read_stream = ObsMetadata.read_info_file(info_file, self.infofiles_path)
        obj = Filter.dynamic_class_constructor(ObsMetadata(read_stream['filter']), "")
        
        if self.test:    
        #   self.assertTrue(isinstance(obj, obj.type), f'Object {info_file} is not a {obj.type} filter')    
            
           if re.match("test---attributes", str(info_file)): # compare with expected result
               self._filter_compare(info_file, obj, test_expected_result)
        
        
        if self.verbose:
           print(f'Filter test for: {info_file}: PASSED')
        
        if self.print_output:    
           print(obj)
           
    
    def _filter_compare(self, info_file, filter, expected_result):
        """
        Test a created filter object against an expected result to make sure all values are right
        """
        ftype = filter.type 
        read_dict = vars(filter)
        
        # Remove notes and extras
        if read_dict.get('notes', None) == []:
            read_dict.pop('notes')
        if read_dict.get('extras', None) == None:
            read_dict.pop('extras')
        
        self.assertEqual(read_dict,expected_result.get(ftype, None), 
f" File: '{info_file}'. Computed result: {read_dict} and expected result: {expected_result.get(ftype, None)} are different")
        
    
    def _test_PZ_conditionals(self):
        """
        Test all the conditionals in the PZ filter, and the function to calculate the normalization factor.
        """
        obj = Filter.dynamic_class_constructor(ObsMetadata({'type': 'PolesZeros', 
                                         'transfer_function_type' : 'LAPLACE (RADIANS/SECOND)',
                                         'zeros': [[0.3, 0.2],],
                                         'poles': [[0.546, 0.191], [4.40e4, 0.000],],
                                         'normalization_frequency' : 1.,
                                         'normalization_factor' : None,
                                          }), None, None, None)
        self.assertIn(obj.transfer_function_type, ['LAPLACE (RADIANS/SECOND)', 'LAPLACE (HERTZ)', 'DIGITAL (Z-TRANSFORM)'], 
                                              f'transfer function type wrong in test case for PZ {obj.transfer_function_type}')
        self.assertEqual(obj.normalization_factor,  44188.013594177224, 
                f'object normalization factor in test case for PZ {obj.normalization_factor} in PZ test is different from 44188.013594177224')
        print("1", end=" ")
        
        obj = Filter.dynamic_class_constructor(ObsMetadata({'type': 'PolesZeros', 
                                         'transfer_function_type' : 'LAPLACE (HERTZ)',
                                         'zeros': [[0.3, 0.2],],
                                         'poles': [[0.546, 0.191], [4.40e4, 0.000],],
                                         'normalization_frequency' : 1.,
                                         'normalization_factor' : None,
                                          }), None, None, None)
        self.assertIn(obj.transfer_function_type, ['LAPLACE (RADIANS/SECOND)', 'LAPLACE (HERTZ)', 'DIGITAL (Z-TRANSFORM)'], 
                                              f'transfer function type wrong in test case for PZ {obj.transfer_function_type}')
        self.assertEqual(obj.normalization_factor,  50262.70428857582, 
                    f'object normalization factor in test case for PZ {obj.normalization_factor} is different from 50262.70428857582')
        print("2", end=" ")
        
        obj = Filter.dynamic_class_constructor(ObsMetadata({'type': 'PolesZeros', 
                                         'transfer_function_type' : 'LAPLACE (RADIANS/SECOND)',
                                         'zeros': [[0.3, 0.2],],
                                         'poles': [[0.546, 0.191], [4.40e4, 0.000],],
                                         'normalization_frequency' : 120.,
                                         'normalization_factor' : None,
                                          }), None, None, None)
                                          
        self.assertEqual(obj.normalization_factor, 44006.99311749303, f'{obj.normalization_factor} in test case for PZ is different from 44006.99311749303')
        print("3", end=" ")
        
        obj = Filter.dynamic_class_constructor(ObsMetadata({'type': 'PolesZeros', 
                                         'transfer_function_type' : 'LAPLACE (RADIANS/SECOND)',
                                         'zeros': [],
                                         'poles': [],
                                         'normalization_frequency' : 1.,
                                         'normalization_factor' : None,
                                          }), None, None, None)
        
        self.assertEqual(obj.normalization_factor, 1., f'{obj.normalization_factor} is different from 1.')
        print("4", end=" ")
        
        obj = Filter.dynamic_class_constructor(ObsMetadata({'type': 'PolesZeros', 
                                         'transfer_function_type' : 'DIGITAL (Z-TRANSFORM)',
                                         'zeros': [[0.3, 0.2],],
                                         'poles': [[0.546, 0.191], [4.40e4, 0.000],],
                                         'normalization_frequency' : 1.,
                                         'normalization_factor' : None,
                                          }), None, None, None)
        
        self.assertIn(obj.transfer_function_type, ['LAPLACE (RADIANS/SECOND)', 'LAPLACE (HERTZ)', 'DIGITAL (Z-TRANSFORM)'], 
                                              f'transfer function type wrong in test case for PZ {obj.transfer_function_type}')
        self.assertEqual(obj.normalization_factor,  None, 
                    f'object normalization factor {obj.normalization_factor} is different from None')
        print("5", end=" ")
        
        obj = Filter.dynamic_class_constructor(ObsMetadata({'type': 'PolesZeros', 
                                         'transfer_function_type' : 'LAPLACE (RADIANS/SECOND)',
                                         'zeros': [[0.3, 0.2],],
                                         'poles': [[0.546, 0.191], [4.40e4, 0.000],],
                                         'normalization_frequency' : None,
                                         'normalization_factor' : None,
                                          }), None, None, None)
        
        self.assertEqual(obj.normalization_factor, None, f'{obj.normalization_factor} is different from None')
        print("6: Should have returned error", end=" ")
        
        
    
    
    def _test_all_stage_types(self):
        """
        Test reading and converting to obspy a stage file with each filter type. 
        This is the first time obspy conversion occurs so make sure it's done right
        Only one example stage for each type. If you wish to test all stage files,
        use test_all_responses(). If you wish to test individual files use test_state_with_XXX(file)
        with file as stripped file name. 
        File must exist in obsinfo/Information_files/instrumentation/YYY/responses
        """
        self.test_stage('TI_ADS1281_FIR1.stage.yaml')
        self.test_stage('SIO-LDEO_DPG_5018_calibrated.stage.yaml')
        self.test_stage('test-with-coeff.stage.yaml')
        self.test_stage('test-with-response-list.stage.yaml')
        
    def test_all_stages(self):
        """
        Test all information files in each responses subdirectory.
        """
        for dir in self.infofiles_path.datapath_list:
            files_in_validate_dir = Path(dir).joinpath(
                "*rs", # includes sensors, preamplifiers and dataloggers
                "responses/*.yaml"
                )
            
            filelist = glob.glob(str(files_in_validate_dir))
            
            for file in filelist:
                self._test_stage(file)
        
        
    def _test_stage(self, file):
        """
        Test stage according to filter type
        """
        
        info_file_dict = ObsMetadata.read_info_file(file, self.infofiles_path)
           
        stage_from_info_file = Stage(ObsMetadata(info_file_dict['stage']))
        
        obspy_result = stage_from_info_file.to_obspy()
        self._test_common_attributes(stage_from_info_file, obspy_result)
        
        if self.test:
            if isinstance(filter, FIR):
                self.assertEqual(stage_from_info_file.filter.symmetry, obspy_result._symmetry)
                for info_file_coeff in stage_from_info_file.filter.coefficients:
                    for obspy_coeff in obspy_result.decimation_correction:
                        self.assertEqual(info_file_coeff / 512, obspy_coeff (f))
            elif isinstance(filter, PolesZeros):
                self.assertEqual(stage_from_info_file.filter.transfer_function_type, obspy_result.pz_transfer_function_type)
                self.assertEqual(stage_from_info_file.filter.normalization_frequency, obspy_result.normalization_frequency)
                self.assertEqual(stage_from_info_file.filter.normalization_factor, obspy_result.normalization_factor)
                self.assertEqual(stage_from_info_file.filter.zeros, obspy_result.zeros)
                self.assertEqual(stage_from_info_file.filter.poles, obspy_result.poles)
            elif isinstance(filter, ResponseList):
                self.assertEqual(stage_from_info_file.filter.response_list, obspy_result.response_list_elements)
            elif isinstance(filter, Coefficients):
                self.test_common_attributes(stage_from_info_file, obspy_result)
                self.assertEqual(stage_from_info_file.filter.transfer_function_type, obspy_result.cf_transfer_function_type)
                self.assertEqual(stage_from_info_file.filter.numerator_coefficients, obspy_result.numerator)
                self.assertEqual(stage_from_info_file.filter.denominator_coefficients, obspy_result.denominator)                
                
        if self.verbose:
            print(f'Stage test for: {file}: PASSED')
            
        if self.print_output:
            print(obj)
            if self.level == "all":
                print(s.filter) 
    
    def _test_common_attributes(self, stage_from_info_file, obspy_result):
        self.assertEqual(stage_from_info_file.name, obspy_result.name)
        self.assertEqual(stage_from_info_file.description, obspy_result.description)
        self.assertEqual(stage_from_info_file.input_units, obspy_result.input_units)
        self.assertEqual(stage_from_info_file.output_units, obspy_result.output_units)
        self.assertEqual(stage_from_info_file.input_units_description, obspy_result.input_units_description)
        self.assertEqual(stage_from_info_file.output_units_description, obspy_result.output_units_description)
        self.assertEqual(stage_from_info_file.gain, obspy_result.stage_gain)
        self.assertEqual(stage_from_info_file.gain_frequency, obspy_result.stage_gain_frequency)
        self.assertEqual(stage_from_info_file.decimation_factor, obspy_result.decimation_factor)
        self.assertEqual(stage_from_info_file.filter.offset, obspy_result.decimation_offset)
        self.assertEqual(stage_from_info_file.delay, obspy_result.decimation_delay)
        self.assertEqual(stage_from_info_file.correction, obspy_result.decimation_correction)
         
         
    def _test_response_stage_addition(self):
        """
        Test reading and combining response_stages.
        """
        read_info_A = ObsMetadata.read_info_file(PurePath(self.infofiles_path).joinpath( 
            'sensors',
            'responses',
            'Trillium_T240_SN400-singlesided_theoretical.stage.yaml'))
        read_info_B = ObsMetadata.read_info_file(PurePath(self.infofiles_path).joinpath(
            'dataloggers',
            'responses',
            'TexasInstruments_ADS1281_100sps-linear_theoretical.response_stages.yaml'))
        stages_A = ResponseStages(read_info_A['response']['stages'])
        stages_B = ResponseStages(read_info_B['response']['stages'])
        stages = stages_A + stages_B
        
        
    def test_all_components(self):
        """
        Test all information files in each responses subdirectory.
        """
        components_list = [
            "sensors",
            "preamplifiers",
            "dataloggers"
            ]
        
        for dir in self.infofiles_path.datapath_list:
            for comp in components_list:
                files_in_validate_dir = Path(dir).joinpath(
                    comp,
                    "*.yaml"  # includes sensors, preamplifiers and dataloggers
                    )
                
                filelist = glob.glob(str(files_in_validate_dir))
                
                for file in filelist:
                    info_file_dict = ObsMetadata.read_info_file(file, self.infofiles_path)
        
                    #OJO: no configuraton passed from above. No delay_correction either.
                    obj = InstrumentComponent.dynamic_class_constructor(comp[:-1], info_file_dict)
                    
                    #if self.test:
                    #    self.assertTrue(isinstance(obj. comp[:-1]))
                        
                    if self.verbose:   
                       print(f'{file}: PASSED')
                       
                    if self.print_output:
                        PrintObs.print_component(obj, self.level)
                        
    
    def test_all_instrumentations(self):
        """
        Test all information files in each responses subdirectory.
        """
        for dir in self.infofiles_path.datapath_list:
            files_in_validate_dir = Path(dir).joinpath(
                "instrumentation/*.yaml" # includes sensors, preamplifiers and dataloggers
                )
            
            filelist = glob.glob(str(files_in_validate_dir))
            
            for file in filelist:
                self._test_instrumentation(file)
                    
         
    def _test_instrumentation(self, file):
        """
        Test reading instrumentation.
        """
        dict = ObsMetadata.read_info_file(file, self.infofiles_path)
        
        start_date = end_date = UTCDateTime(0)
        
        location_dict = {
                           "position": {"lon": 0., "lat": 0., "elev": 0.},
                           "base":  { "depth.m": 0.,
                                "geology": "unknown",
                                "vault": "Sea floor",
                                "uncertainties.m": {"lon": 0., "lat": 0., "elev": 0.},
                                "localisation_method": "Sea surface release point"
                               }
                        }
        
        location = Location(location_dict)
        locations={"00": location}
        
        print(f'Processing "{file}"')
        obj = Instrumentation(ObsMetadata(dict['instrumentation']), locations, start_date, end_date, {})
     
        if self.verbose:   
           print(f'{file}: PASSED')
        
        if self.print_output:
            PrintObs.print_instrumentation(obj, self.level)
            
        
    def _test_station(self, file_name):
        """
        Test reading a station.
        """
        info_dict = ObsMetadata.read_info_file(file_name, self.infofiles_path)
        
        info_dict = info_dict['stations']
        key = list(info_dict.keys())[0]
        value = ObsMetadata(list(info_dict.values())[0])
        
        obj = Station(key, value)
        
        if self.verbose:
           print(f'Station test for: {file_name}: PASSED')
           
        if self.print_output:
            PrintObs.print_station(obj, self.level)
            
    
    def test_all_networks(self):
        """
        Test all information files in each responses subdirectory.
        """
        for dir in self.infofiles_path.datapath_list:
            files_in_validate_dir = Path(dir).joinpath(
                "network/*.yaml" # includes sensors, preamplifiers and dataloggers
                )
            
            filelist = glob.glob(str(files_in_validate_dir))
            
            for file in filelist:
                self._test_network(file)
                    
             
    def _test_network(self, file_name):
        """
        Test reading a network.
        """
        info_dict = ObsMetadata.read_info_file(file_name, self.infofiles_path)
        
        net_dict = info_dict.get('network',None)
        if not net_dict:
            return 
        
        if self.verbose:
            print(f'Processing network file: {file_name}')
        obj = Network(ObsMetadata(net_dict))
        
        if self.verbose:
           print(f'Network test for: {file_name}: PASSED')
           
        if self.print_output:
            PrintObs.print_network(obj, self.level)
       
    
def run_suite_yaml():
    """
    Create and run test suite for jsonref
    """
    
    suite_yaml = unittest.TestSuite()
    suite_yaml.addTests(unittest.makeSuite(JsonRefTest))
     
    result = unittest.TextTestRunner(verbosity=1).run(suite_yaml)
    
    report_result_summary(result)
    
    return suite_yaml
    
def run_suite_info_files():
    """
    Create test suites for information files  
    """
    def suite():
        suite_info_files = unittest.TestSuite()
        suite_info_files.addTest(TestObsinfo('test_all_filters'))
        suite_info_files.addTest(TestObsinfo('test_all_stages'))
        suite_info_files.addTest(TestObsinfo('test_all_components'))
        suite_info_files.addTest(TestObsinfo('test_all_instrumentations'))
        suite_info_files.addTest(TestObsinfo('test_all_networks'))
        
        return suite_info_files
        
    
    result = unittest.TextTestRunner(verbosity=1).run(suite())
      
    report_result_summary(result)
        
    
def report_result_summary(result):
    """
    Report a summary of errors and failures
    """
    
    n_errors = len(result.errors)
    n_failures = len(result.failures)

    if n_errors or n_failures:
        print('\n\nSummary: %d errors and %d failures reported\n'%\
            (n_errors, n_failures)) 
           

def print_obs():
    """
    Print an information file
    """

    level, print_output, debug, input_filename = retrieve_arguments()
    val = TestObsinfo()
    val.setUp(verbose=False, test=False, print_output=print_output, level=level) 
    dp = Datapath() 
        
    try:               
        
        type = ObsMetadata.get_information_file_type(input_filename)
         
        print(f'Printing {type} file: {input_filename}')
            
        if type == "filter":
            val._test_filter(val.infofiles_path.build_datapath(input_filename))
        elif type == "stage":
            val._test_stage(val.infofiles_path.build_datapath(input_filename))
        elif type == "datalogger":
            val._test_datalogger(val.infofiles_path.build_datapath(input_filename))
        elif type == "preamplifier":
            val._test_preamplifier(val.infofiles_path.build_datapath(input_filename))
        elif type == "sensor":
            val._test_sensor(val.infofiles_path.build_datapath(input_filename))
        elif type == "instrumentation":
            val._test_instrumentation(val.infofiles_path.build_datapath(input_filename))
        elif type == "station":
            val._test_station(val.infofiles_path.build_datapath(input_filename))
        elif type == "network":
            val._test_network(val.infofiles_path.build_datapath(input_filename))
                       
    except TypeError:
        if debug:
            raise
        print("Illegal format: fields may be missing or with wrong format in input file.")
    except (KeyError, IndexError):
        if debug:
            raise
        print("Illegal value in dictionary key or list index")
    except ValueError:
        if debug:
            raise
        print("An illegal value was detected")
    except (IOError, OSError, LookupError):
        if debug:
            raise
        print("File could not be opened or read")
    except FileNotFoundError:
        if debug:
            raise
        print("File could not be found")
    except JSONDecodeError:
        if debug:
            raise
        print("File format is not recognized as either JSON or YAML. Check syntax.") 


def retrieve_arguments():
    
    options_dict = {
                       "level" : "l",
                       "debug": "d",
                       "help" : "h",
                     }
    
    input_filename = output_filename = None
    level ="all"
    print_output = True 
    debug = False
    
    long_option = re.compile("^[\-][\-][a-zA-Z_]+$")
    short_option = re.compile("^[\-][a-zA-Z]$")
    possible_options = re.compile("^[ldh]+$")
    
    input_filename = ""
    
    option = None
    skip_next_arg = False
    
    for arg in sys.argv[1:]:
        
        if skip_next_arg:
            skip_next_arg = False
            continue
      
        if re.match(long_option, arg):  
            option = options_dict.get(arg[2:])
        elif not arg[0] == "-":
            input_filename = arg
            continue  
        else:
            option = arg[1:]
              
        if not re.match(possible_options, option):
            s = f'Unrecognized option in command line: -{option}\n'
            s += usage()
            raise ValueError(s)
        
        for opt in option:
    
            if opt == "l":
                if len(option) == 1:
                    level = sys.argv[sys.argv.index("-l" if "-l" in sys.argv else "--level")+1]
                    skip_next_arg = True
                else:
                    warnings.warn('-l option should stand alone and be followed by a depth number')
                    break
            elif opt == "d":
                debug = True
            elif opt == "h": 
                print(usage())
                sys.exit()
             
    return (level, print_output, debug, input_filename)   

def usage():
    s = f'Usage: {sys.argv[0]} -pdh [-l <level name>]<filename>\n'
    s += f'Where:\n'
    s += f'      -l or --level: prints up to <level name>, where <level name> is:\n'
    s += f'                     all\n'
    s += f'                     stage\n'
    s += f'                     component\n'
    s += f'                     instrumentation\n'
    s += f'                     channel\n'
    s += f'                     station\n'
    s += f'                     network\n'
    s += f'      -d or --debug: do not catch exceptions to show error trace.\n'
    s += f'      -h or --help: prints this message\n'
    
    return s



"""
Generate scripts needed to go from basic miniSEED to data center ready
"""
import obsinfo
from obsinfo.network import network as oi_network
import os.path
import sys
from obspy.core import UTCDateTime

SEPARATOR_LINE = "\n# " + 60 * "=" + "\n"


################################################################################
def process_script(
    station,
    station_dir,
    distrib_dir="/opt/sdpchain",
    input_dir="miniseed_basic",
    corrected_dir="miniseed_corrected",
    extra_commands=None,
    include_header=True,
    SDS_uncorr_dir="SDS_uncorrected",
    SDS_corr_dir="SDS_corrected",
    SDS_combined_dir="SDS_combined",
):
    """Writes OBS data processing script using SDPCHAIN software

        station:          an obsinfo.station object
        station_dir:      base directory for the station data
        input_dir:        directory beneath station_dir for input (basic)
                          miniseed data ['miniseed_basic']
        corrected_dir:    directory beneath station_dir for output (corrected)
                          miniseed data ['miniseed_corrected']
        SDS_corr_dir:     directory beneath station_dir in which to write
                          SDS structure of corrected data (ideally
                          ../SOMETHING if ms2sds could write all to the
                          same directory)
        SDS_uncorr_dir:   directory beneath station_dir in which to write
                          SDS structure of uncorrected data (ideally
                          ../SOMETHING if ms2sds could write all to the
                          same directory)
        include_header:   whether or not to include the bash script header
                          ('#!/bin/bash') at the top of the script [True]
        distrib_dir:      Base directory of sdpchain distribution ['/opt/sdpchain']

        The sequence of commands is:
            1: optional proprietary format steps (proprietary format -> basic miniseed, separate)
            2: optional extra_steps (any cleanup needed for the basic
                miniseed data, should either overwrite the existing data or
                remove the original files so that subsequent steps only see the
                cleaned data)
            3: ms2sds on basic miniseed data
            4: leap-second corrections, if necessary
            5: msdrift (creates drift-corrected miniseed)

    """
    leap_corr_dir = "miniseed_leap_corrected"

    s = ""
    if include_header:
        s += __header(station.code)
    s += __setup_variables(distrib_dir, station_dir)
    if extra_commands:
        s += __extra_command_steps(extra_commands)
    s += __ms2sds_script(station, input_dir, SDS_uncorr_dir)
    t = __leap_second_script(
        station.clock_corrections.get("leapseconds", None), input_dir, leap_corr_dir
    )
    if t:
        s += t
        input_dir = leap_corr_dir
    s += __msdrift_script(input_dir, corrected_dir, station.clock_corrections)
    s += __force_quality_script(corrected_dir, "Q")
    s += __ms2sds_script(station, corrected_dir, SDS_corr_dir)
    s += __combine_sds_script(station, SDS_corr_dir, SDS_uncorr_dir, SDS_combined_dir)

    return s


############################################################################
def __header(station_name):

    s = "#!/bin/bash\n"
    s += SEPARATOR_LINE + f'echo "Working on station {station_name}"' + SEPARATOR_LINE

    return s


############################################################################
def __setup_variables(distrib_dir, station_dir):

    s = SEPARATOR_LINE + "# SDPCHAIN STEPS" + SEPARATOR_LINE
    s += "#  - Set up paths\n"
    s += f"STATION_DIR={station_dir}\n"
    s += f"MSDRIFT_EXEC={os.path.join(distrib_dir,'bin','msdrift')}\n"
    s += f"MSDRIFT_CONFIG={os.path.join(distrib_dir,'config','msdrift.properties')}\n"
    s += f"MS2SDS_EXEC={os.path.join(distrib_dir,'bin','ms2sds')}\n"
    s += f"MS2SDS_CONFIG={os.path.join(distrib_dir,'config','ms2sds.properties')}\n"
    s += f"SDPPROCESS_EXEC={os.path.join(distrib_dir,'bin','sdp-process')}\n"
    s += f"MSMOD_EXEC={os.path.join('/opt/iris','bin','msmod')}\n"
    s += "\n"

    return s


############################################################################
def __extra_command_steps(extra_commands):
    """
    Write extra command lines, embedded in sdp-process

    Input:
        extra_commands: list of strings containing extra commands
    """
    s = SEPARATOR_LINE
    s = s + "# - EXTRA COMMANDS\n"
    if not isa(extra_commands, "list"):
        error("extra_commands is not a list")
    for cmd_line in extra_commands:
        s = s + '$SDPPROCESS_EXEC --cmd="{cmd_line}"\n'
    return s


############################################################################
def __ms2sds_script(station, in_path, out_path):

    """
    Writes the ms2sds lines of the script
    """
    sta = station.code
    net = station.network_code

    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running MS2SDS: MAKE SDS ARCHIVE"\n'
    s += f'echo "{"-"*60}"\n'

    s += f'in_dir="{in_path}"\n'
    s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Collect input filenames\n"
    s += "command cd $STATION_DIR/$in_dir\n"
    s += "mseedfiles=$(ls *.mseed)\n"
    s += "command cd -\n"
    s += 'echo "mseedfiles=" $mseedfiles\n'

    s += "# - Run executable\n"
    s += "$MS2SDS_EXEC $mseedfiles -d $STATION_DIR -i $in_dir -o $out_dir "
    s += f'--network "{net}" --station "{sta}" -a SDS -p $MS2SDS_CONFIG\n'
    s += "\n"

    return s


############################################################################
def __leap_second_script(leapseconds, in_dir, out_dir):
    """
    Create leap-second correction text

    Inputs:
        leapseconds: list of dictionaries from network information file
    """
    if not leapseconds:
        return ""

    s = f'echo "{"-"*60}"\n'
    s += 'echo "LEAPSECOND CORRECTIONS"\n'
    s += f'echo "{"-"*60}"\n'

    s += f"in_dir={in_dir}\n"
    s += f"out_dir={out_dir}\n"

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Copy files to output directory\n"
    s += "cp $STATION_DIR/$in_dir/*.mseed $STATION_DIR/$out_dir\n"

    for leapsecond in leapseconds:
        if leapsecond["corrected_in_basic_miniseed"]:
            s += "# LEAP SECOND AT {} ALREADY CORRECTED IN BASIC MINISEED, DOING NOTHING\n".format(
                leapsecond["time"]
            )
            return s
        temp = leapsecond["time"].split("T")
        d = UTCDateTime(temp[0])
        leap_time = d.strftime("%Y,%j,") + temp[1].rstrip("Z")
        s += 'echo ""\n'
        s += f'echo "{"="*60}"\n'
        s += 'echo "Running LEAPSECOND correction"\n'
        s += f'echo "{"-"*60}"\n'
        if leapsecond["type"] == "+":
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Shifting one second BACKWARDS after positive leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --timeshift -1 -ts {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Marking the record containing the positive leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --actflags 4,1 -tsc {leap_time} -tec {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
        elif leapsecond["type"] == "-":
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Shifting one second FORWARDS after negative leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --timeshift +1 -ts {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Marking the record containing the negative leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --actflags 5,1 -tsc {leap_time} -tec {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
        else:
            s += 'ERROR: leapsecond type "{}" is neither "+" nor "-"\n'.format(
                leapsecond["type"]
            )
            sys.exit(2)
    return s


############################################################################
def __msdrift_script(in_path, out_path, clock_corrs):
    """
    Write msdrift lines of the script

    Inputs:
        in_path
        out_path
        clock_corrs
    """
    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running MSDRIFT: CORRECT LINEAR CLOCK DRIFT"\n'
    s += f'echo "{"-"*60}"\n'

    s += f'in_dir="{in_path}"\n'
    s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Collect input filenames\n"
    s += f"command cd $STATION_DIR/$in_dir\n"
    s += "mseedfiles=$(ls *.mseed)\n"
    s += "command cd -\n"
    s += 'echo "mseedfiles=" $mseedfiles\n'

    if "linear_drift" in clock_corrs:
        lin_corr = clock_corrs["linear_drift"]
        s += "# - Run executable\n"
        s += f'START_REFR="{str(lin_corr["start_sync_reference"]).rstrip("Z")}"\n'
        s += f'START_INST="{str(lin_corr["start_sync_instrument"]).rstrip("Z")}"\n'
        s += f'END_REFR="{str(lin_corr["end_sync_reference"]).rstrip("Z")}"\n'
        s += f'END_INST="{str(lin_corr["end_sync_instrument"]).rstrip("Z")}"\n'
        s += f"$MSDRIFT_EXEC $mseedfiles -d $STATION_DIR -i $in_dir -o $out_dir "
        s += f'-m "%E.%S.00.%C.%Y.%D.%T.mseed:%E.%S.00.%C.%Y.%D.%T_driftcorr.mseed" '
        s += f'-s "$START_REFR;$START_INST" '
        s += f'-e "$END_REFR;$END_INST" '
        # s += f'-c "comment.txt" '
        s += f"-p $MSDRIFT_CONFIG\n"
        s += "\n"
    else:
        while lin_corr in clock_corrs["linear_drifts"]:
            s += +SEPARATOR + LINE
            s += +"ERROR, CANT YET APPLY MULTIPLE TIME CORRECTIONS (SHOULD CHANGE\n"
            s += (
                +"MSDRIFT TO ONLY WRITE GIVEN TIME RANGE AND BE ABLE TO APPEND TO EXISTING FILE?)\n"
            )
            s += +SEPARATOR + LINE
    return s


############################################################################
def __force_quality_script(in_path, quality="Q"):
    """
    Force data quality to Q

    Inputs:
        in_path: directory where mseed files are found
        quality: new quality code
    """
    s = f'echo "{"-"*60}"\n'
    s += 'echo "Forcing data quality to Q"\n'
    s += f'echo "{"-"*60}"\n'
    s += f'$SDPPROCESS_EXEC -d $STATION_DIR -c="Forcing data quality to Q" --cmd="$MSMOD_EXEC --quality Q -i {in_path}/*.mseed"\n'
    s += "\n"

    return s


############################################################################
def __combine_sds_script(
    station, SDS_corrected_dir, SDS_uncorrected_dir, SDS_combined_dir
):
    """
    Combine corrected and uncorrected SDS directories/files

    Corrected must be before uncorrected so that Arclink will extract it by default
    Uses the corrected directory as the template: will not process uncorrected files
    that don't have corresponding corrected files

    Creates no process-steps because there are so many individual calls to msmod (should we write an sdp-sds_combine script?)

    Inputs:
        station:             station object
        SDS_corrected_dir:   name of SDS directory containing corrected data
        SDS_uncorrected_dir: name of SDS directory containing uncorrected data
        SDS_combined_dir:    name of SDS directory to put combined data into
    """
    s = f'echo "{"-"*60}"\n'
    s += 'echo "Combining corrected and uncorrected SDS datafiles"\n'
    s += f'echo "{"-"*60}"\n'
    s += f"mkdir -p $STATION_DIR/{SDS_combined_dir}\n"
    s += f"command cd $STATION_DIR/{SDS_corrected_dir}\n"
    s += f"years=????\n"
    s += f"command cd -\n"
    s += f"for y in $years ; do\n"
    # I HAD TO ADD "SDSs" to handle the fact that ms2sds makes an "SDS" directory inside the output directory
    # s += f'    for d in $STATION_DIR/{SDS_corrected_dir}/$y/{station.network_code}/{station.code}/*.D ; do\n'
    # s +=  '        d_sub=${{d#$STATION_DIR/{}/}}\n'.format(SDS_corrected_dir)
    s += f"    for d in $STATION_DIR/{SDS_corrected_dir}/SDS/$y/{station.network_code}/{station.code}/*.D ; do\n"
    s += "        d_sub=${{d#$STATION_DIR/{}/SDS/}}\n".format(SDS_corrected_dir)
    s += f"        if [[ ! -d $STATION_DIR/{SDS_combined_dir}/$d_sub ]] ; then\n"
    s += f'            echo "Creating and filling combined SDS subdirectory $d_sub"\n'
    s += f'            eval "mkdir -p $STATION_DIR/{SDS_combined_dir}/$d_sub"\n'
    s += f"            for f in $d/*; do\n"
    s += "                f_sub=${{f#$STATION_DIR/{}/SDS/}}\n".format(SDS_corrected_dir)
    s += f"                cat $STATION_DIR/{SDS_corrected_dir}/SDS/$f_sub "
    s += f"$STATION_DIR/{SDS_uncorrected_dir}/SDS/$f_sub > "
    s += f"$STATION_DIR/{SDS_combined_dir}/$f_sub\n"
    s += f"            done\n"
    s += f"        else\n"
    s += f'            echo "Combined SDS subdirectory $d_sub already exists, skipping..."\n'
    s += f"        fi\n"
    s += f"    done\n"
    s += f"done\n"
    s += "\n"

    return s


################################################################################
def _console_script(argv=None):
    """
    Console-level processing script

    requires SDPCHAIN programs msdrift and ms2sds, and IRIS program msmod

    Currently usese 'SDS_corrected' and 'SDS_uncorrected' as default SDS directories
    Would be better to use ../SDS_[un]corrected so that all data are together,
    but ms2sds is not yet capable of putting data in an existing SDS directory

    """
    from argparse import ArgumentParser

    parser = ArgumentParser(
        prog="obsinfo-make_process_scripts_SDPCHAIN", description=__doc__
    )
    parser.add_argument("network_file", help="Network information file")
    parser.add_argument("station_data_path", help="Base path containing station data")
    parser.add_argument(
        "distrib_dir", help="Path to SDPCHAIN software", default="/opt/sdpchain"
    )
    parser.add_argument(
        "-i",
        "--input_dir",
        default="miniseed_basic",
        help="subdirectory of station_data_path/{STATION}/ containing input *.mseed files",
    )
    parser.add_argument(
        "-o",
        "--corrected_dir",
        default="miniseed_corrected",
        help="subdirectory of station_data_path/{STATION}/ to put corrected *.mseed files",
    )
    parser.add_argument(
        "--SDS_corr_dir",
        default="SDS_corrected",
        help="subdirectory of station_data_path/{STATION}/ for SDS structure of corrected data",
    )
    parser.add_argument(
        "--SDS_uncorr_dir",
        default="SDS_uncorrected",
        help="subdirectory of station_data_path/{STATION}/ for SDS structure of uncorrected data",
    )
    parser.add_argument(
        "--suffix", default="_SDPCHAIN", help="suffix for script filename"
    )
    parser.add_argument(
        "--append", action="store_true", help="append to existing script file"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="increase output verbosiy"
    )
    parser.add_argument(
        "--no_header", action="store_true", help="do not include file header"
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="run silently")
    args = parser.parse_args()

    # READ IN NETWORK INFORMATION
    if not args.quiet:
        print(f"Creating SDPCHAIN process scripts, ", end="")
    network = oi_network(args.network_file)
    if not args.quiet:
        print(f"network {network.network_info.code}, stations ", end="")
        if args.verbose:
            print("")

    first_time = True
    for name, station in network.stations.items():
        if not args.quiet:
            if args.verbose:
                print(f"\t{name}", end="")
            else:
                if first_time:
                    print(f"{name}", end="")
                else:
                    print(f", {name}", end="")
        station_dir = os.path.join(args.station_data_path, name)
        script = process_script(
            station,
            station_dir,
            distrib_dir=args.distrib_dir,
            input_dir=args.input_dir,
            corrected_dir=args.corrected_dir,
            SDS_uncorr_dir=args.SDS_uncorr_dir,
            SDS_corr_dir=args.SDS_corr_dir,
            include_header=not args.no_header,
        )
        fname = "process_" + name + args.suffix + ".sh"
        if args.verbose:
            print(f" ... writing file {fname}", flush=True)
        if args.append:
            write_mode = "a"
        else:
            write_mode = "w"
        with open(fname, write_mode) as f:
            f.write(script)
            f.close()
        first_time = False
    if not args.verbose and not args.quiet:
        print("")



"""
Instrumentation and Instrument classes

nomenclature:
    An "Instrument" (measurement instrument) records one physical parameter
    An "Instrumentation" combines one or more measurement instruments
"""
# Standard library modules
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
# Non-standard modules
from obspy.core.inventory.network import Network as obspy_Network
import obspy.core.util.obspy_types as obspy_types
from obspy.core.inventory.util import (Person, Comment, PhoneNumber)
from obspy.core.inventory.util import Operator as obspy_Operator
from obspy.core.utcdatetime import UTCDateTime
# obsinfo modules
from ..obsMetadata.obsmetadata import (ObsMetadata)
from ..network.station import (Station, Location)
from ..instrumentation.instrumentation import (Operator)


class Network(ObsMetadata):
    """
    One or more Instruments. Part of an obspy/StationXML Station
    """
    def __init__(self, attributes_dict=None):
        """
        Constructor
        """
        
        if not attributes_dict:
            raise TypeError('No network attributes')
        
        self.campaign_ref = attributes_dict.get("campaing_ref_name", None)
        network_info = attributes_dict.get("network_info", None)
        
        self.fdsn_code = network_info.get("code", None)
        self.fdsn_name = network_info.get("name", None)        
        self.start_date = network_info.get("start_date", 0)
        self.end_date = network_info.get("end_date", 0)
        self.description = network_info.get("description", None)
        
        self.restricted_status = attributes_dict.get("restricted_status", None)
        
        self.operator = Operator(attributes_dict.get("operator", None))  
        
        stations = attributes_dict.get("stations", None)              
        self.stations = [Station(k, v) for k, v in stations.items()]
        
        self.comments = attributes_dict.get("comments", [])
        self.extras = [str(k) + ": " + str(v) for k, v in (attributes_dict.get('extras', {})).items()]
        self.convert_notes_and_extras_to_obspy() 
        
        self.obspy_network = self.to_obspy()     
          

    def __repr__(self):
        s = f'Network(Campaign: {self.campaign_ref}, FDSN Code: {self.fdsn_code}, FDSN Name: {self.fdsn_code}, '
        s += f'Start date: {self.start_date}, End date: {self.end_date}, Description: {self.description}, '
        s += f'{len(self.stations)} stations)'

        return s
    
    def to_obspy(self):
        """
         Convert network object to obspy object
        """
          
        if self.operator:
            
            mails = [self.operator.email] if self.operator.email else []
            phones = [PhoneNumber(self.operator.area_code, 
                                  self.operator.phone_number,
                                  self.operator.country_code)] if self.operator.phone_number else []
            names = [self.operator.contact] if self.operator.contact else []
            agencies=[self.operator.reference_name] if self.operator.reference_name else []
            person = Person(names=names, 
                            agencies=agencies, 
                            emails=mails, 
                            phones=phones)
                   
            operator = obspy_Operator(
                agency=[self.operator.full_name],  
                contacts=[person], 
                website=self.operator.website)
        else:
            operator = None
                      
        stations_number = len(self.stations)
        start_date = UTCDateTime(self.start_date if self.start_date else 0)
        end_date = UTCDateTime(self.end_date if self.end_date else 0)
        comments = [Comment(s) for s in self.comments]
         
        self.obspy_network = obspy_Network(code=self.fdsn_code, 
                                     stations=  [st.obspy_station for st in self.stations], 
                                     total_number_of_stations=stations_number, 
                                     selected_number_of_stations=stations_number, 
                                     description=self.description, 
                                     comments=comments,
                                     start_date=start_date, 
                                     end_date=end_date, 
                                     restricted_status=self.restricted_status, 
                                     alternate_code=None, 
                                     historical_code=None, 
                                     data_availability=None, 
                                     identifiers=None, 
                                     operators=[operator], 
                                     source_id=None)
        
        for st in self.stations:  #complete stations
            st.operators = [operator]
            
        return self.obspy_network
    
    def convert_notes_and_extras_to_obspy(self):
        """
        Convert info file notes and extras to XML comments
        """
    
        self.comments += ['EXTRA ATTRIBUTES (for documentation only): '] + self.extras
    
    
    


import sys
import os
import re
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
from pathlib import Path, PurePath
from json.decoder import JSONDecodeError

import obspy
from obspy.core.inventory import Inventory,  Station, Channel, Site
from obspy.core.inventory import Network as obspy_Network
from obspy.clients.nrl import NRL

from ..network import (Network)
from ..obsMetadata.obsmetadata import (ObsMetadata)
from ..misc.discoveryfiles import (Datapath)
from ..print_version import main as print_version 
import obsinfo



# We'll first create all the various objects. These strongly follow the
# hierarchy of StationXML files.

# Succesful treatment 
EXIT_SUCCESS = 0
# File could not be processed, generic reasons 
EXIT_FAILURE = 1

# Reference: https://www.freebsd.org/cgi/man.cgi?query=sysexits&manpath=FreeBSD+4.3-RELEASE

# bad CLI parameters 
EXIT_USAGE = 64
#bad file content (input file : Network metadata file) 
EXIT_DATAERR = 65
# file not found (input file : Network metadata file) 
EXIT_NOINPUT = 66 
#A service is unavailable (This can occur if a support program or file does not exist) 
EXIT_UNAVAILABLE = 69 

# (technical/internal error)
# An internal software error has been detected 
EXIT_SOFTWARE = 70

#can't create file (report file, output file)
EXIT_CANTCREAT = 73 
#An error occurred while doing I/O on some file */
EXIT_IOERR = 74 

#Something was found in an unconfigured or misconfigured state */
EXIT_CONFIG = 78 


def main():
    
    dp = Datapath()
    # create list of directories to search for files    
    verbose, quiet, test, validate, debug, remote, input_filename, output_filename, schemapath = retrieve_arguments(dp)
    
    if verbose:
        print(f'Using OBSINFO_DATAPATH: {dp.datapath_list}')    
        
    try:               
                         
        file = Path(input_filename).name    
       
        if validate:
            if verbose:
                print(f'Validating network file: {file}')
                
            ret = ObsMetadata().validate(schemapath,  str(input_filename), "yaml", 
                                         "network", verbose, "network", False)
                    
        info_dict = ObsMetadata.read_info_file(input_filename, dp, remote)
        
        net_dict = info_dict.get('network', None)
        if not net_dict:
            return 
                
        if verbose:
            print(f'Processing network file: {file}')
        
        obj = Network(ObsMetadata(net_dict))
                  
        if verbose:
           print(f'Network file parsed successfully for: {file}')
           
        networks=[obj.obspy_network]
        if not isinstance(obj.obspy_network, obspy_Network):
            print("Not a network object")
            
        if not quiet:
           print(obj.obspy_network)
        
            
        version = os.environ.get('OBSINFO_VERSION')
        
        author = get_first_author(info_dict)
    
        inv = Inventory(
                networks,
                module=version,
                module_uri="https://gitlab.com/resif/obsinfo",
                sender=author,
                source="ObsPy")
    
        if not test: # Generate Stationxml file
            stem_name = Path(file).stem      # remove .yaml
            stem_name = Path(stem_name).stem #Remove .network
            #stem_name = re.split("\\.", file)  #OJO Cambiar esto porque puede haber puntos antes
            output_filename = stem_name + ".station.xml"
            stationxml=inv.write(output_filename, format="stationxml", validate=False)
         
        if not quiet:
           print(f'StationXML file created successfully: {output_filename}')        
               
    except TypeError:
        if debug:
            raise
        print("Illegal format: fields may be missing or with wrong format in input file.")
        sys.exit(EXIT_DATAERR)
    except (KeyError, IndexError):
        if debug:
            raise
        print("Illegal value in dictionary key or list index")
        sys.exit(EXIT_SOFTWARE)
    except ValueError:
        if debug:
            raise
        print("An illegal value was detected")
        sys.exit(EXIT_DATAERR)
    except FileNotFoundError:
        if debug:
            raise
        print("File could not be found")
        sys.exit(EXIT_NOINPUT)
    except JSONDecodeError:
        print("File and/or subfiles have an illegal format. Probably indentation or missing quotes/parentheses/brackets")
        if debug:
            raise 
        sys.exit(EXIT_DATAERR)
    except (IOError, OSError, LookupError):
        if debug:
            raise
        print("File could not be opened or read")
        sys.exit(EXIT_UNAVAILABLE)
    except:
        sys.exit(EXIT_FAILURE)
    
  
def retrieve_arguments(datapath):
    
    options_dict = {
                       "output": "o",
                       "verbose" : "v",
                       "version" : "v",
                       "quiet" : "q",
                       "debug" : "d",
                       "test" : "t",
                       "validate" : "l",
                       "remote_discovery" : "r",
                       "help" : "h",
                     }
    
    input_filename = output_filename = None
    verbose = quiet = test = False
    validate = debug = remote = False
    
    skip_next_arg = False
    path_exists = False
            
    long_option = re.compile("^[\-][\-][a-zA-Z_]+$")
    short_option = re.compile("^[\-][a-zA-Z]$")
    possible_options = re.compile("^[Vvqtldorh]+$")
    
    input_filename = ""
    
    option = None
    
    for arg in sys.argv[1:]:

        if skip_next_arg:
            skip_next_arg = False
            continue
        
        if re.match(long_option, arg):  
            option = options_dict.get(arg[2:])
        elif not arg[0] == "-":
            input_filename = arg
            continue  
        else:
            option = arg[1:]
        
        if not re.match(possible_options, option):
            s = f'Unrecognized option in command line: -{option}\n'
            s += usage()
            sys.exit(EXIT_USAGE)
        
        for opt in option:
    
            if opt == "o":
                if len(option) == 1:
                    output_filename = sys.argv[sys.argv.index("-o" if "-o" in sys.argv else "--output")+1]
                    skip_next_arg = True
                else:
                    warnings.warn('-o option should stand alone and be followed by a filename')
                    break
            elif opt == "v":
                verbose = True
            elif opt == "q":
                quiet = True
            elif opt == "d":
                debug = True  
            elif opt == "t":
                test = True
            elif opt == "l":
                validate = True
            elif opt == "V":
                print_version()
                sys.exit(EXIT_SUCCESS)
            elif opt == "r":
                remote = True
            elif opt == "h": 
                print(usage())
                sys.exit(EXIT_USAGE)
            
        skip_next_arg = False
    
    # schemas must always be installed under obsinfo/data/schemas   
    schemapath = Path(obsinfo.__file__).parent.joinpath('data', 'schemas')
    
    if not input_filename:
        print("No input filename specified")
        sys.exit(EXIT_USAGE) 
    
    input_filename = str(datapath.build_datapath(arg) if remote else Path(os.getcwd()).joinpath(input_filename))
            
    return (verbose, quiet, test, validate, debug, remote, input_filename, output_filename, schemapath)   

def get_first_author(info_dict):
    """
    Get info file first author info and return a string
    """
    rev = info_dict.get('revision', None)
    if not rev:
        return ""
   
    authors = rev.get('authors', None)

    if authors:
        author = authors[0].get('first_name', "") + " " + authors[0].get('last_name', "") \
           + ", " + authors[0].get('institution', "") +  " Email: " + authors[0].get('email', "") + "  Phones: "
        for phn in authors[0].get('phones', ""):
            author += phn + " "
    
    return author 
        

def usage():
    s = f'Usage: {sys.argv[0]} -vqdtlrdh  [-o <filename>] [-i] <filename>\n'
    s += f'Where:\n'
    s += f'      -v or --verbose: prints processing progression\n'
    s += f'      -q or --quiet: silences a human readable summary of processed information file\n'
    s += f'      -d or --debug: turns on the trace of exceptions\n'
    s += f'      -t or --test: enters test mode, produces no output\n'
    s += f'      -l or --validate: validate the YAML or JSON format of the information file. Used when file syntax is trusted\n'
    s += f'      -r or --assume input filename is discovered through OBSINFO_DATAPATH environment variable. Does not affect treatment of $ref in info files\n'
    s += f'      -h or --help: prints this message\n'
    s += f'      -o or --output: names the output file. Default is station.xml\n'
    
    
    return s
    
if __name__ == '__main__':
    main()


"""
Filter class and subclasses
"""
# Standard library modules
import math as m
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)

from scipy._lib.doccer import extend_notes_in_docstring

# Non-standard modules
# import obspy.core.inventory.response as obspy_response


class Filter(object):
    """
    Filter superclass
    """
    def __init__(self, type="PolesZeros", offset=0):
        """
        Constructor
        """
        self.type = type
        self.offset = offset
        
    @staticmethod
    def dynamic_class_constructor(attributes_dict, channel_modif={}, selected_config={}, stage_id="-1"):
        """
        Creates an appropriate Filter subclass from an attributes_dict
        stage_id is stage_name if it exists
        """
        
        if attributes_dict == None:
            raise TypeError("No attributes in filter")
        
        if "type" not in attributes_dict:
            raise TypeError('No "type" specified for filter in stage #{stage_id}')
        else:
            filter_type = attributes_dict.get_configured_element('type', 
                                                           channel_modif, 
                                                           selected_config, 
                                                           None)
            if filter_type == 'PolesZeros':
                obj = PolesZeros.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            elif filter_type == 'FIR':
                obj = FIR.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            elif filter_type == 'Coefficients':
                obj = Coefficients.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            elif filter_type == 'ResponseList':
                obj = ResponseList.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            elif filter_type == 'ADConversion':
                obj = AD_Conversion.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            elif filter_type == 'Analog':
                obj = Analog.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            elif filter_type == 'Digital':
                obj = Digital.dynamic_class_constructor(filter_type, attributes_dict, selected_config, stage_id)
            else:
                raise TypeError(f'Unknown Filter type: "{filter_type}" in stage #{stage_id}')
                
        return obj
    
        
class PolesZeros(Filter):
    """
    PolesZeros Filter
    """
    def __init__(self, filter_type, transfer_function_type='LAPLACE (RADIANS/SECOND)', 
                 poles=[], zeros=[],
                 normalization_frequency=1., 
                 normalization_factor=None,
                 offset=0,
                 stage_id=-1):
        """
        poles and zeros should be lists of complex numbers
        """
        if transfer_function_type not in ['LAPLACE (RADIANS/SECOND)','LAPLACE (HERTZ)','DIGITAL (Z-TRANSFORM)']:
            raise TypeError(f'Illegal transfer_function_type in PolesZeros: "{transfer_function_type}" in stage #{stage_id}')
            
        self.transfer_function_type = transfer_function_type
        self.poles = poles
        self.zeros = zeros
        
        self.normalization_frequency = normalization_frequency
        if normalization_frequency and normalization_factor:
            self.normalization_factor = normalization_factor
        elif filter_type != 'Analog':     #norm frequency doesn't make sense in analog      
            self.normalization_factor = self.calc_normalization_factor(stage_id)
        else:
            self.normalization_factor = 1.0

        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}, stage_id=-1):
        """
        Create PolesZeros instance from an attributes_dict
        """
        obj = cls(filter_type,
                attributes_dict.get_configured_element('transfer_function_type', 
                                                  channel_modif, 
                                                  selected_config, 
                                                  'LAPLACE (RADIANS/SECOND)'),
                [(float(x[0]) + 1j*float(x[1])) for x in attributes_dict.get_configured_element('poles',
                                                         channel_modif, 
                                                         selected_config, 
                                                         [])],
                [(float(x[0]) + 1j*float(x[1])) for x in attributes_dict.get_configured_element('zeros',
                                                         channel_modif, 
                                                         selected_config,
                                                         [])],
                attributes_dict.get_configured_element('normalization_frequency', 
                                                 channel_modif, 
                                                 selected_config, 
                                                 1.),
                attributes_dict.get_configured_element('normalization_factor', 
                                                 channel_modif, 
                                                 selected_config, 
                                                 None),
                attributes_dict.get_configured_element('offset', 
                                                 channel_modif, 
                                                 selected_config,
                                                 0),
                stage_id
              )
        return obj

    def __repr__(self):
        """
        String representation of object
        """
        s = f'          PolesZeros(Poles={self.poles}, Zeros={self.zeros}, '
        s += f'Normalization Frequency={self.normalization_frequency:g}, '
        s += f'Normalization Factor={self.normalization_factor:g})'
        return s

    def calc_normalization_factor(self, stage_id=-1, debug=False):
        """
        Calculate the normalization factor for give poles-zeros

        The norm factor A0 is calculated such that
                           sequence_product_over_n(s - zero_n)
                A0 * abs(------------------------------------------) === 1
                           sequence_product_over_m(s - pole_m)
        for s_f=i*2pi*f if the transfer function is in radians
                i*f     if the transfer function is in Hertz
        """
        if not self.normalization_frequency:
            raise ValueError(f'No normalization frequency for PZ filter in stage #{stage_id}')

        A0 = 1.0 + (1j * 0.0)
        if self.transfer_function_type == "LAPLACE (HERTZ)":
            s = 1j * self.normalization_frequency
        elif self.transfer_function_type == "LAPLACE (RADIANS/SECOND)":
            s = 1j * 2 * m.pi * self.normalization_frequency
        else:
            warnings.warn("Don't know how to calculate normalization factor "
                  "for z-transform poles and zeros!") #Warning msg
            return None
        for p in self.poles:
            A0 *= (s - p)
        for z in self.zeros:
            A0 /= (s - z)

        if debug:
            print("poles={}, zeros={}, s={:g}, A0={:g}".format(
                self.poles, self.zeros, s, A0))

        A0 = abs(A0)
        return A0


class FIR(Filter):
    """
    FIR Filter
    """
    def __init__(self, filter_type, symmetry, coefficients, coefficient_divisor, offset=0, stage_id=-1):
        
        self.symmetry = symmetry
        if symmetry not in ['ODD', 'EVEN', 'NONE']:
            raise TypeError(f'Illegal FIR symmetry: "{symmetry} in stage #{stage_id}"')
        
        #Validate coefficients
        sum_coeff = 0
        coeff_cnt = 0
        for coeff in coefficients:
            sum_coeff += coeff
            coeff_cnt += 1
        
        coeff = sum_coeff / coefficient_divisor
        coeff = round(coeff, 2) #check up to two decimal places  
        if coeff != 1 and coeff != 0: #the last conditional means there is at least one coeff
            warnings.warn(f'Coefficient sum "{coeff}" not equal to one, num. coefficient: "{coeff_cnt}" in stage #{stage_id}'), 
            
        self.coefficients = coefficients
        self.coefficient_divisor = coefficient_divisor
        
        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}, stage_id=-1):
        """
        Create FIR instance from an attributes_dict
        """
        
        offset = attributes_dict.get_configured_element('offset', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   0)
        if not offset:
            raise TypeError(f'No offset in FIR filter')
        
        obj = cls(filter_type,
                  attributes_dict.get_configured_element('symmetry', 
                                                   channel_modif, 
                                                   selected_config,
                                                   None), #Default will cause an error in init
                  attributes_dict.get_configured_element('coefficients', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   []),
                  attributes_dict.get_configured_element('coefficient_divisor', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   1.),
                  offset)
        
        return obj

    def __repr__(self):
        """
        String representation of object
        """
        s = f'          FIR("Symmetry={self.symmetry}",'
        s += f' Coefficients={self.coefficients}, Divisor={self.coefficient_divisor})'
        return s


class Coefficients(Filter):
    """
    Coefficients Filter
    """
    def __init__(self, filter_type, transfer_function_type, numerator_coefficients,
                 denominator_coefficients, offset=0):
        if transfer_function_type not in ["ANALOG (RADIANS/SECOND)",
                                          "ANALOG (HERTZ)",
                                          "DIGITAL"]:
            raise TypeError('Illegal transfer function type: "{transfer_function_type}" in stage #{stage_id}')

        self.transfer_function_type = transfer_function_type
        self.numerator_coefficients = numerator_coefficients
        self.denominator_coefficients = denominator_coefficients
        
        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}, stage_id=-1):
        """
        Create Coefficients instance from an attributes_dict
        """
        obj = cls(filter_type, 
                  attributes_dict.get_configured_element('transfer_function_type', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   'DIGITAL'),
                  attributes_dict.get_configured_element('numerator_coefficients', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   []),
                  attributes_dict.get_configured_element('denominator_coefficients', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   []),
                  attributes_dict.get_configured_element('offset', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   0)
                  )
        return obj

    def __repr__(self):
        s = f'          Coefficients("Transfer Function Type={self.transfer_function_type}", '
        s += f'Numerator={self.numerator_coefficients}, '
        s += f'Denominator={self.denominator_coefficients})'
        return s


class ResponseList(Filter):
    """
    ResponseList Filter
    """
    def __init__(self, filter_type, response_list, offset=0):
        self.response_list = response_list
        
        super().__init__(filter_type, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}, stage_id=-1):
        """
        Create Response List instance from an attributes_dict
        """
        obj = cls(filter_type, 
                  attributes_dict.get_configured_element('elements', channel_modif, selected_config, []), 
                  attributes_dict.get_configured_element('offset', channel_modif, selected_config, 0))
        
        return obj

    def __repr__(self):
        return f'          ResponseList("{self.response_list}")'


class Analog(PolesZeros):
    """
    Analog Filter (Flat PolesZeros filter)
    """
    def __init__(self, filter_type, offset=0): 
        
        self.poles = []
        self.zeros = []
        self.normalization_frequency = 1.
        self.normalization_factor = None
        
        super().__init__(filter_type, "LAPLACE (RADIANS/SECOND)", 
                         self.poles, self.zeros, self.normalization_frequency, self.normalization_factor, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}):
        """
        Create Analog instance from an attributes_dict
        """
        obj = cls(filter_type, attributes_dict.get_configured_element('offset', 
                                                                channel_modif, 
                                                                selected_config, 
                                                                0))
        return obj

    def __repr__(self):
        return '          Analog()'


class Digital(Coefficients):
    """
    Digital Filter (Flat Coefficients filter)
    """
    def __init__(self, filter_type, offset=0):
        self.transfer_function_type = 'DIGITAL'
        self.numerator_coefficients = [1.0]
        self.denominator_coefficients = []

        super().__init__(filter_type, "DIGITAL", self.numerator_coefficients, self.denominator_coefficients, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}, stage_id=-1):
        """
        Create Digital instance from an attributes_dict
        """
        obj = cls(filter_type, attributes_dict.get_configured_element('offset', 
                                                                channel_modif, 
                                                                selected_config, 
                                                                0))
        return obj

    def __repr__(self):
        return '          Digital()'


class AD_Conversion(Coefficients):
    """
    AD_Conversion Filter (Flat Coefficients filter)
    """
    def __init__(self, filter_type, input_full_scale, output_full_scale, offset=0):
        self.transfer_function_type = 'DIGITAL'
        self.numerator_coefficients = [1.0]
        self.denominator_coefficients = []
        self.input_full_scale = input_full_scale
        self.output_full_scale = output_full_scale
        
        super().__init__(filter_type, "DIGITAL", 
                         self.numerator_coefficients, self.denominator_coefficients, offset)

    @classmethod
    def dynamic_class_constructor(cls, filter_type, attributes_dict, channel_modif={}, selected_config={}, stage_id=-1):
        """
        Create AD_Conversion instance from an attributes_dict
        """
        obj = cls(filter_type, 
                  attributes_dict.get_configured_element('input_full_scale', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   None),
                  attributes_dict.get_configured_element('output_full_scale', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   None), 
                  attributes_dict.get_configured_element('offset', 
                                                   channel_modif, 
                                                   selected_config, 
                                                   0))
        return obj

    def __repr__(self):
        s = f'          ADConversion(Input Full Scale={self.input_full_scale:g}, '
        s += f'Output Full Scale={self.output_full_scale})'
        return s



"""
jsonref with YAML reading added

copied directly from jsonref v0.2, with added routines _yaml_load
and _yaml_loads replacing json.load and json.loads
Added/modified lines are marked "# WCC"
"""
import functools
import os
import json
import yaml # WCC
import operator
import re
import sys
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
import copy
from json.decoder import JSONDecodeError
from pathlib import Path

from ..misc.remoteGitLab import gitLabFile

from ..misc.discoveryfiles import Datapath

try:
    from collections.abc import Mapping, MutableMapping, Sequence
except ImportError:
    from collections import Mapping, MutableMapping, Sequence

PY3 = sys.version_info[0] >= 3

from urllib import parse as urlparse
from urllib.parse import unquote
from urllib.request import urlopen

unicode = str
basestring = str
iteritems = operator.methodcaller("items")

try:
    # If requests >=1.0 is available, we will use it
    import requests

    if not callable(requests.Response.json):
        requests = None
except ImportError:
    requests = None

from proxytypes import LazyProxy, Proxy

__version__ = "0.2"


class JsonRefError(Exception):
    def __init__(self, message, reference, uri="", base_uri="", path=(), cause=None):
        self.message = message
        self.reference = reference
        self.uri = uri
        self.base_uri = base_uri
        self.path = list(path)
        self.cause = self.__cause__ = cause

    def __repr__(self):
        return "<%s: %r>" % (self.__class__.__name__, self.message)

    def __str__(self):
        return str(self.message)


class JsonRef(LazyProxy):
    """
    A lazy loading proxy to the dereferenced data pointed to by a JSON
    Reference object.

    """

    __notproxied__ = ("__reference__",)

    @classmethod
    def replace_refs(cls, obj, _recursive=False, **kwargs):
        """
        Returns a deep copy of `obj` with all contained JSON reference objects
        replaced with :class:`JsonRef` instances.

        :param obj: If this is a JSON reference object, a :class:`JsonRef`
            instance will be created. If `obj` is not a JSON reference object,
            a deep copy of it will be created with all contained JSON
            reference objects replaced by :class:`JsonRef` instances
        :param base_uri: URI to resolve relative references against
        :param loader: Callable that takes a URI and returns the parsed JSON
            (defaults to global ``jsonloader``, a :class:`JsonLoader` instance)
        :param jsonschema: Flag to turn on `JSON Schema mode
            <http://json-schema.org/latest/json-schema-core.html#anchor25>`_.
            'id' keyword changes the `base_uri` for references contained within
            the object
        :param load_on_repr: If set to ``False``, :func:`repr` call on a
            :class:`JsonRef` object will not cause the reference to be loaded
            if it hasn't already. (defaults to ``True``)

        """
               
        store = kwargs.setdefault("_store", _URIDict())
        base_uri, frag = urlparse.urldefrag(kwargs.get("base_uri", ""))
        
        store_uri = None  # If this does not get set, we won't store the result
        if not frag and not _recursive:
            store_uri = base_uri
        try:
            if kwargs.get("jsonschema") and isinstance(obj["id"], basestring): # It's a schema
                kwargs["base_uri"] = urlparse.urljoin( 
                    kwargs.get("base_uri", ""), obj["id"]
                )
                store_uri = kwargs["base_uri"]
        except (TypeError, LookupError):
            pass

        try:
            if not isinstance(obj["$ref"], basestring): #It's a $ref but not a string
                raise TypeError
        except (TypeError, LookupError):
            pass
        else:
            return cls(obj, **kwargs) #__init__ and Return JsonRef object (content of $ref) if a string

        # If our obj was not a json reference object, iterate through it,
        # replacing children with JsonRefs
        kwargs["_recursive"] = True
        path = list(kwargs.pop("_path", ()))
        if isinstance(obj, Mapping):
            obj = type(obj)( #Calls __init__ for the relevant object type
                (k, cls.replace_refs(v, _path=path + [k], **kwargs))
                for k, v in iteritems(obj)
            )
        elif isinstance(obj, Sequence) and not isinstance(obj, basestring):
            obj = type(obj)( #Calls __init__ for the relevant object type
                cls.replace_refs(v, _path=path + [i], **kwargs)
                for i, v in enumerate(obj)
            )
        #else obj is a "leaf" object (not a list, not a directory) and thus not replaceable, we simply return it
        
        if store_uri is not None:
            store[store_uri] = obj
            
        return obj

    def __init__(
        self,
        refobj,
        base_uri="",
        loader=None,
        jsonschema=False,
        load_on_repr=True,
        _path=(),
        _store=None,
        datapath=None,
    ):
        if not isinstance(refobj.get("$ref"), basestring):
            raise ValueError("Not a valid json reference object: %s" % refobj)

        self.__reference__ = refobj
        self.base_uri = base_uri 
        self.loader = loader or jsonloader
        self.jsonschema = jsonschema
        self.load_on_repr = load_on_repr
        self.path = list(_path)
        self.store = _store  # Use the same object to be shared with children
        self.datapath = datapath
        if self.store is None:
            self.store = _URIDict()

    @property
    def _ref_kwargs(self):
        return dict(
            base_uri=self.base_uri,
            loader=self.loader,
            jsonschema=self.jsonschema,
            load_on_repr=self.load_on_repr,
            _path=self.path,
            _store=self.store,
            datapath = self.datapath,
        )

    @property
    def full_uri(self):
        
        kwargs = self._ref_kwargs
        
        if kwargs['jsonschema']:
            return urlparse.urljoin(self.base_uri, self.__reference__["$ref"]) 
        else:        
            dp = kwargs["datapath"]
            if not dp:
                print(f'Error in datapath in full_uri, reference: {self.__reference__["$ref"]}')
                raise ValueError
            base_uri = Path(dp.build_datapath(self.__reference__["$ref"]))
            
            self.base_uri = unquote(urlparse.urlsplit(str(base_uri)).path if gitLabFile.isRemote(str(base_uri)) \
                                      else base_uri.as_uri())
            return(self.base_uri) 
    
        
    def callback(self):
        uri, fragment = urlparse.urldefrag(self.full_uri)
        # If we already looked this up, return a reference to the same object
        if uri in self.store:
            result = self.resolve_pointer(self.store[uri], fragment)
        else:
            # Remote ref
            try:              
                base_doc = self.loader(uri)
            except Exception as e:
                # self._error("%s: %s" % (e.__class__.__name__, unicode(e)), cause=e)
                # WCC
                print(e)
                return None
                

            kwargs = self._ref_kwargs
            kwargs["base_uri"] = uri 
            base_doc = JsonRef.replace_refs(base_doc, **kwargs)
            result = self.resolve_pointer(base_doc, fragment)
            
        if hasattr(result, "__subject__"):
            # TODO: Circular ref detection
            result = result.__subject__
        return result

    def resolve_pointer(self, document, pointer):
        """
        Resolve a json pointer ``pointer`` within the referenced ``document``.

        :argument document: the referent document
        :argument str pointer: a json pointer URI fragment to resolve within it

        """
        # Do only split at single forward slashes which are not prefixed by a caret
        parts = re.split(r"(?<!\^)/", unquote(pointer.lstrip("/"))) if pointer else []

        for part in parts:
            # Restore escaped slashes and carets
            replacements = {r"^/": r"/", r"^^": r"^"}
            part = re.sub(
                "|".join(re.escape(key) for key in replacements.keys()),
                lambda k: replacements[k.group(0)],
                part,
            )
            if isinstance(document, Sequence):
                # Try to turn an array index to an int
                try:
                    part = int(part)
                except ValueError:
                    pass
            try:
                document = document[part]
            except (TypeError, LookupError) as e:
                self._error("Unresolvable JSON pointer: %r" % pointer, cause=e)
        return document

    def _error(self, message, cause=None):
        raise JsonRefError(
            message,
            self.__reference__,
            uri=self.full_uri,
            base_uri=self.base_uri,
            path=self.path,
            cause=cause,
        )

    def __repr__(self):
        if hasattr(self, "cache") or self.load_on_repr:
            return repr(self.__subject__)
        return "JsonRef(%r)" % self.__reference__


class _URIDict(MutableMapping):
    """
    Dictionary which uses normalized URIs as keys.

    """

    def normalize(self, uri):
        return urlparse.urlsplit(uri).geturl()

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.store.update(*args, **kwargs)

    def __getitem__(self, uri):
        return self.store[self.normalize(uri)]

    def __setitem__(self, uri, value):
        self.store[self.normalize(uri)] = value

    def __delitem__(self, uri):
        del self.store[self.normalize(uri)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __repr__(self):
        return repr(self.store)


class JsonLoader(object):
    """
    Provides a callable which takes a URI, and returns the loaded JSON referred
    to by that URI. Uses :mod:`requests` if available for HTTP URIs, and falls
    back to :mod:`urllib`. By default it keeps a cache of previously loaded
    documents.

    :param store: A pre-populated dictionary matching URIs to loaded JSON
        documents
    :param cache_results: If this is set to false, the internal cache of
        loaded JSON documents is not used

    """

    def __init__(self, store=(), cache_results=True):
        self.store = _URIDict(store)
        self.cache_results = cache_results

    def __call__(self, uri, **kwargs):
        """
        Return the loaded JSON referred to by `uri`

        :param uri: The URI of the JSON document to load
        :param kwargs: Keyword arguments passed to :func:`json.loads`

        """
        if uri in self.store:
            return self.store[uri]
        else:
            result = self.get_remote_json(uri, **kwargs)
            if self.cache_results:
                self.store[uri] = result
            return result

    def get_remote_json(self, uri, **kwargs):
        scheme = urlparse.urlsplit(uri).scheme
        path = urlparse.urlsplit(uri).path

        # Open locally or remotely according to scheme   
        
        if scheme == 'file':
            try:
                with open(path, "rt") as fp:
                    strm = fp.read()
            except FileNotFoundError:
                print(f'File not found: {path}')
                raise FileNotFoundError(path)
            except (IOError, OSError):
                print(f'Input/Output error with file: {path}')
                raise
        else:
            strm = gitLabFile.get_gitlab_file(path)
          
        result = _yaml_loads(strm, **kwargs)   # LFA
        
        return result


jsonloader = JsonLoader()
# jsonloader = YAMLLoader() # WCC


def load(fp, base_uri="", loader=None, jsonschema=False, load_on_repr=True, datapath=None,
         **kwargs):
    """
    Drop in replacement for :func:`json.load`, where JSON references are
    proxied to their referent data.

    :param fp: File-like object containing JSON document
    :param kwargs: This function takes any of the keyword arguments from
        :meth:`JsonRef.replace_refs`. Any other keyword arguments will be
        passed to :func:`json.load`

    """
    if loader is None:
        loader = functools.partial(jsonloader, **kwargs)
        
    # Convert encoded characters
    base_uri = unquote(base_uri)
    
    return JsonRef.replace_refs(
        _yaml_load(fp, **kwargs), # WCC
        base_uri=base_uri,
        loader=loader,
        jsonschema=jsonschema,
        load_on_repr=load_on_repr,
        datapath=datapath, #LFA. Datapath is the list of paths where files can potentialy be found
    )


def loads(s, base_uri="", loader=None, jsonschema=False, 
          load_on_repr=True, datapath=None, recursive=True, **kwargs):
    """
    Drop in replacement for :func:`json.loads`, where JSON references are
    proxied to their referent data.

    :param s: String containing JSON document
    :param kwargs: This function takes any of the keyword arguments from
        :meth:`JsonRef.replace_refs`. Any other keyword arguments will be passed to
        :func:`json.loads`

    """
    
    if loader is None:
        loader = functools.partial(jsonloader, **kwargs)
    
    dic = _yaml_loads(s, **kwargs) if isinstance(s, str) else s
        
    if recursive:
        return JsonRef.replace_refs(
            dic,  # WCC et LFA
            base_uri=base_uri,
            loader=loader,
            jsonschema=jsonschema,
            load_on_repr=load_on_repr,
            datapath=datapath,
        )
    else:
        return dic


def load_uri(uri, base_uri=None, loader=None, jsonschema=False, load_on_repr=True):
    """
    Load JSON data from ``uri`` with JSON references proxied to their referent
    data.

    :param uri: URI to fetch the JSON from
    :param kwargs: This function takes any of the keyword arguments from
        :meth:`JsonRef.replace_refs`

    """

    if loader is None:
        loader = jsonloader
    if base_uri is None: 
        base_uri = uri

    return JsonRef.replace_refs(
        loader(uri),
        base_uri=base_uri,
        loader=loader,
        jsonschema=jsonschema,
        load_on_repr=load_on_repr,
        datapath=datapath,
    )

# WCC ADDED BEGIN
# Method read_json_yaml_ref  in ObsMetadata uses load, not loads. Load loads a file.
def _yaml_load(fp, **kwargs):
    """ Call {yaml,json}.load according to file type """
    try:
        return json.load(fp, **kwargs)
    # except JSONDecodeError:
    except:
        fp.seek(0)
        try:
            return yaml.safe_load(fp)
        # except ScannerError:
        except:
            warnings.warn(f'file {fp.name} is neither JSON nor YAML.')
            raise JSONDecodeError("File and/or subfiles have an illegal format. Probably indentation or missing quotes/parentheses/brackets",
                                  fp.name, 0)
    return None

# Method validate in ObsMetadata uses loads, not load. loads loads a string, not a file.
def _yaml_loads(s, **kwargs):
    """ Call {yaml,json}.loads according to file type """
    a = None
    
    kw = copy.deepcopy(kwargs) #copy to make sure you don alter the original kwargs
    kw.pop('base_uri', None)
    kw.pop('datapath', None)
    
    if s[:3] == '---':
        try:
            a = yaml.safe_load(s, **kw)
        except:
            try:
                a = json.loads(s, **kwargs)
            except:
                raise #JSONDecodeError(f'string is neither JSON nor YAML:\n', "file")
      
    else:
        try:
            a = json.loads(s, **kwargs)
        except:
            try:
                a = yaml.safe_load(s, **kwargs)
            except:
                raise #JSONDecodeError('string is neither JSON nor YAML')
                
            
    return a
# WCC ADDED END

def dump(obj, fp, **kwargs):
    """
    Serialize `obj`, which may contain :class:`JsonRef` objects, as a JSON
    formatted stream to file-like `fp`. `JsonRef` objects will be dumped as the
    original reference object they were created from.

    :param obj: Object to serialize
    :param fp: File-like to output JSON string
    :param kwargs: Keyword arguments are the same as to :func:`json.dump`

    """
    # Strangely, json.dumps does not use the custom serialization from our
    # encoder on python 2.7+. Instead just write json.dumps output to a file.
    fp.write(dumps(obj, **kwargs))


def dumps(obj, **kwargs):
    """
    Serialize `obj`, which may contain :class:`JsonRef` objects, to a JSON
    formatted string. `JsonRef` objects will be dumped as the original
    reference object they were created from.

    :param obj: Object to serialize
    :param kwargs: Keyword arguments are the same as to :func:`json.dumps`

    """
    kwargs["cls"] = _ref_encoder_factory(kwargs.get("cls", json.JSONEncoder))
    return json.dumps(obj, **kwargs)


def _ref_encoder_factory(cls):
    class JSONRefEncoder(cls):
        def default(self, o):
            if hasattr(o, "__reference__"):
                return o.__reference__
            return super(JSONRefEncoder, cls).default(o)

        # Python 2.6 doesn't work with the default method
        def _iterencode(self, o, *args, **kwargs):
            if hasattr(o, "__reference__"):
                o = o.__reference__
            return super(JSONRefEncoder, self)._iterencode(o, *args, **kwargs)

        # Pypy doesn't work with either of the other methods
        def _encode(self, o, *args, **kwargs):
            if hasattr(o, "__reference__"):
                o = o.__reference__
            return super(JSONRefEncoder, self)._encode(o, *args, **kwargs)

    return JSONRefEncoder



"""
Response Stage class
"""
# Standard library modules
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
import re

# Non-standard modules
from obspy.core.inventory.response import (PolesZerosResponseStage,
                                           FIRResponseStage,
                                           CoefficientsTypeResponseStage,
                                           ResponseListResponseStage,
                                           Response)
from obspy.core.inventory.response import Response as obspy_Response

import obspy.core.util.obspy_types as obspy_types

# Local modules
from ..obsMetadata.obsmetadata import (ObsMetadata)
from .filter import (Filter, PolesZeros, FIR, Coefficients, ResponseList,
                     Analog, Digital, AD_Conversion)


class ResponseStages():
    """
    ReponseStages class

    An ordered list of Stages. Method can be invoked with attribute_list as a single element of class 
    Stage or as a list of them. 
    """
    def __init__(self, attribute_list, channel_modif={}, selected_config={}, delay_correction=None):
        
        if attribute_list == None:
            raise TypeError(f'No stages found in information file')
        elif not isinstance(attribute_list, list): #Not a list, single element
            self.stages = Stage(ObsMetadata(attribute_list), 
                                 channel_modif,
                                 selected_config, 
                                 delay_correction, 
                                 0) #last argument is position in list

        else:
            self.stages = [Stage(ObsMetadata(s), 
                                 channel_modif,
                                 selected_config, 
                                 delay_correction, 
                                 i) for s, i in zip(attribute_list, range(0, len(attribute_list)))] #last argument is position in list

        self.obspy_stages = self.to_obspy()

        #delay, delay corrections and sample rates (if not present for digital stages)  will be calculated for the whole instrumentation
           
        
    def to_obspy(self):
        """
        Return equivalent obspy response class
        """
        if not self.stages:
            return None
        
        obspy_stages = [s.to_obspy() for s in self.stages]
        
        return obspy_stages
                                   
        
    def __repr__(self):

        if self.stages != None:
            return f' Number of response stages: {len(self.stages):d}' 
        else:
            return f'Response_Stages: (empty)'
        

class Stage(object):
    """
    Stage class
    """
    def __init__(self, attributes_dict, channel_modif_list={}, selected_config={}, 
                 delay_correction=None, sequence_number=-1):
        
        if attributes_dict == None:
            return None
        
        channel_modif = self.get_selected_modifications(channel_modif_list, str(sequence_number))
        
        name = attributes_dict.get_configured_element('name', channel_modif, selected_config, '')
            
        self.name = name if name else sequence_number #stage id is either the name or a sequence number

        self.description = attributes_dict.get_configured_element('description', 
                                                            channel_modif, 
                                                            selected_config,
                                                            '')
        self.input_units = attributes_dict.get_configured_element('input_units', 
                                                            channel_modif, 
                                                            selected_config, None).get('name', None)
        
        self.output_units = attributes_dict.get_configured_element('output_units', 
                                                             channel_modif, 
                                                             selected_config, None).get('name', None)
        
        gain_dict = ObsMetadata(attributes_dict.get('gain', {}))
        if gain_dict:
            modif_gain_dict = channel_modif.get('gain', {})
            self.gain = gain_dict.get_configured_element('value', 
                                                         modif_gain_dict, 
                                                         selected_config,
                                                          1.0)             
            self.gain_frequency = gain_dict.get_configured_element('frequency', 
                                                                   modif_gain_dict, 
                                                                   selected_config, 
                                                                   0.0)
        else:
            raise TypeError(f'No gain specified in stage {self.name}')
        
        #In this version filter cannot be changed by configuration
        self.filter = Filter.dynamic_class_constructor(ObsMetadata(attributes_dict.get('filter', None)),
                                                       channel_modif.get('filter', None),
                                                       selected_config, 
                                                       self.name)
        if not self.filter:
            raise TypeError(f'No filter in stage {self.name}')
        
        self.stage_sequence_number = -1
        
        input_dict = input_units_description=attributes_dict.get_configured_element('input_units', 
                                                                                    channel_modif, 
                                                                                    selected_config, 
                                                                                    None)
        if input_dict:
            self.input_units_description = input_dict.get('description', None)
        output_dict = output_units_description=attributes_dict.get_configured_element('output_units', 
                                                                                      channel_modif, 
                                                                                      selected_config, 
                                                                                      None)
        if output_dict:  
            self.output_units_description = output_dict.get('description', None)
        self.input_sample_rate = attributes_dict.get_configured_element('input_sample_rate', 
                                                                  channel_modif, 
                                                                  selected_config, 
                                                                  None) 
        
        
        
        self.delay = attributes_dict.get_configured_element('delay', channel_modif, selected_config, None)
        self.decimation_factor = attributes_dict.get_configured_element('decimation_factor', 
                                                                  channel_modif, 
                                                                  selected_config,
                                                                  1)
        self.correction = delay_correction
        
        #default polarity is positive
        self.polarity = attributes_dict.get_configured_element('polarity', channel_modif, selected_config, 1)
                
        #self.calculate_delay(delay_correction)
        
        self.calibration_date = attributes_dict.get_configured_element('calibration_date', selected_config, None)
        
        self.instrument_sensitivity = None #Overall sensitivity will be calculated using obspy and stored in first stage
        
                  
    @property
    def output_sample_rate(self):
        if self.input_sample_rate and self.decimation_factor:
            return self.input_sample_rate / self.decimation_factor
        else:
            return None

    def calculate_delay(self, delay_correct):
        """
        Calculates delay as a function of object for digital filters if delay not specified
        Otherwise, uses delay as passed in the arguments
        Also applies delay_correction in the following way:
            - If delay_correct == None, correction is not applied and self.correction values are made equal to delay values
            - Otherwise, self.correction will be 0 except for the last stage, which will be correct.value
              but is set elsewhere (in datalogger _init_). 
        """
        
        delay = self.delay
        offset = self.filter.offset
        input_sample_rate = self.input_sample_rate
        
        if delay == None:
            if offset:
                if not input_sample_rate:
                    #Delay is already none, leave it like that
                    warnings.warn(
                        f'Stage delay is impossible to calculate \
                        out of filter offset with an Unspecified input sample rate in stage {self.name}. Setting to "None"')
                else: 
                    self.delay = offset / input_sample_rate
            else:    
                self.delay = None
                       
        else:
            self.delay = delay
               
        self.correction = self.delay if delay_correct == None else 0
    
    def get_selected_modifications(self, modif_dict, key_code):
        """
        Select which modifications apply to a given stage, with the stage number (WITHIN an instrument component) as key cide
        """
        
        default_dict = range_dict = {}
        modif = modif_dict.get(key_code, {})
        
        overlap = False
        for k, v in modif_dict.items():
            if k[0] == "*":
                default_dict = v
            elif k[0][0] == "[":             
                if re.match(k, key_code):
                    if overlap:
                        warnings.warn("There is an overlap in response stage modifications. Taking the first applicable pattern.")
                        break # Will only use first match to avoid conflicts
                    range_dict = v
                    overlap = True                  
                                    
        # Gather all modifications in a single dict
        # Do this in order: particular mods have priority over range specific which has
        # priority over default
        for k, v in range_dict.items():
            if k not in modif:    
               modif[k] = v  
           
        for k, v in default_dict.items():
            if k not in modif:    
               modif[k] = v
        
        return modif


    def __repr__(self):
        s = f'\n     Stage("{self.name}", "{self.description}", '
        s += f'"{self.input_units}", "{self.output_units}", '
        s += f'{self.gain}, {self.gain_frequency:g}, '
        s += f'{type(self.filter)}'
        if not self.stage_sequence_number == -1:
            s += f', stage_sequence_number="{self.stage_sequence_number}"'
        if self.input_units_description:
            s += f', input_units_description="{self.input_units_description}"'
        if self.output_units_description:
            s += f', output_units_description='
            s += f'"{self.output_units_description}"'
        if self.input_sample_rate:
            s += f', input_sample_rate={self.input_sample_rate}'
        
        s += f', decimation_factor={self.decimation_factor}'
        s += f', delay={self.delay}'
        s += f', correction={self.correction}'
        
        if self.calibration_date:
            s += f', calibration_dates={self.calibration_date}'
        s += ')'
        return s

    def to_obspy(self):
        """
        Return equivalent obspy response stage
        Don't change parameter assignment in called meethods, as they are used so
        parameter order is irrelevant- Very handy
        """

        filt = self.filter
        args = (self.stage_sequence_number, self.gain, self.gain_frequency,
                self.input_units, self.output_units) # Invariable position arguments for all
        
        if (isinstance(filt, PolesZeros) 
                or isinstance(filt, Analog)):
            if not filt.normalization_frequency:
                filt.normalization_frequency = self.gain_frequency
                
            if filt.type == "PolesZeros": #Only iterate if zeros and poles exist
                PZ_zeros=[obspy_types.ComplexWithUncertainties(
                    t, lower_uncertainty=0.0, upper_uncertainty=0.0)\
                    for t in filt.zeros]
                PZ_poles=[obspy_types.ComplexWithUncertainties(
                    t, lower_uncertainty=0.0, upper_uncertainty=0.0)\
                    for t in filt.poles]
            else:
                PZ_zeros = filt.zeros
                PZ_poles = filt.poles
                
            obspy_stage = PolesZerosResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,                #OJO: check if correction and delay should be modified
                resource_id="",
                # PolesZeros-specific
                pz_transfer_function_type=filt.transfer_function_type,
                normalization_frequency=filt.normalization_frequency,
                zeros=PZ_zeros,
                poles=PZ_poles,
                normalization_factor=filt.normalization_factor
                )
            
        elif isinstance(filt, FIR):
            
            obspy_stage = FIRResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                resource_id="",                
                # FIR-specific
                symmetry=filt.symmetry,
                coefficients=[obspy_types.FloatWithUncertaintiesAndUnit(
                    c / filt.coefficient_divisor) for c in filt.coefficients])
        
        elif (isinstance(filt, Coefficients)
                or isinstance(filt, Digital)
                or isinstance(filt, AD_Conversion)):                
            
            if filt.type == "Coefficients": #Only iterate if zeros and poles exist
                c_numerator=[obspy_types.FloatWithUncertaintiesAndUnit(
                    n, lower_uncertainty=0.0, upper_uncertainty=0.0)\
                    for n in filt.numerator_coefficients]
                c_denominator=[obspy_types.FloatWithUncertaintiesAndUnit(
                    n, lower_uncertainty=0.0, upper_uncertainty=0.0)\
                    for n in filt.denominator_coefficients]
            else:
                c_numerator=filt.numerator_coefficients
                c_denominator=filt.denominator_coefficients
                           
            obspy_stage = CoefficientsTypeResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                resource_id="",
                # CF-specific
                cf_transfer_function_type=filt.transfer_function_type,
                numerator=c_numerator,
                denominator=c_denominator
                )

        elif isinstance(filt, ResponseList):
            
            obspy_stage = ResponseListResponseStage(
                *args,
                name=self.name,
                input_units_description=self.input_units_description,
                output_units_description=self.output_units_description,
                description=self.description,
                decimation_input_sample_rate=self.input_sample_rate,
                decimation_factor=self.decimation_factor,
                decimation_offset=filt.offset,
                decimation_delay=self.delay,
                decimation_correction=self.correction,
                resource_id="",
                # ResponeList-specific
                response_list_elements=filt.response_list)
            
        else:
            raise TypeError(f'Unhandled response stage type in stage #{self.stage_sequence_number}: "{filt.type}"')
        
        return obspy_stage



"""
Instrumentation and Instrument classes

nomenclature:
    An "Instrument" (measurement instrument) records one physical parameter
    A "Channel" is an Instrument + an orientation code and possibly
        starttime, endtime and location code
    An "Instrumentation" combines one or more measurement Channels
"""
# Standard library modules
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
import numpy
import re

# Non-standard modules
from obspy.core.utcdatetime import UTCDateTime
import obspy.core.util.obspy_types as obspy_types
from obspy.core.inventory.response import (Response)
from obspy.core.inventory.response import InstrumentSensitivity\
                                   as obspy_Sensitivity
from obspy.core.inventory.channel import Channel\
                                   as obspy_Channel
from obspy.core.inventory.util import  (Latitude, Longitude, Comment)
# obsinfo modules
from .instrument_component import (InstrumentComponent, Datalogger, Sensor, Preamplifier,
                                   Equipment)
from .response_stages import (ResponseStages)
from ..obsMetadata.obsmetadata import (ObsMetadata)
import obspy.geodetics.base
from obspy.geodetics.base import kilometers2degrees
from numpy import cos

class Instrumentation(object):
    """
    One or more Instruments. Part of an obspy/StationXML Station
    """
    def __init__(self, attributes_dict_or_list, locations, start_date, end_date, channel_modifications={}):
        """
        Constructor
        attributes_dict may contain a configuration_selection for the instrumentation and the corresponding configs  
        for the components: datalogger, preamplifier and sensor

        :param equipment: equipment description
        :type equipment: ~class `Equipment`
        :param channels: list of channels
        :type channels: list of ~class `Channel`
        """
        
        #Syntax checking - Check whether 
        if not attributes_dict_or_list:
            raise TypeError('No instrumentation attributes')
        elif isinstance(attributes_dict_or_list, list): #It's an easy mistake to include a dash in the yaml file
            warnings.warn(f'Instrumentation should be unique, not a list. Removing list embedding.')
            if len(attributes_dict_or_list) > 0:
                attributes_dict = attributes_dict_or_list[0]
            else:
                raise TypeError('No instrumentation attributes')
        else:
            attributes_dict = attributes_dict_or_list
        
        self.operator = Operator(attributes_dict.get('operator', None))   
        self.equipment = Equipment(ObsMetadata(attributes_dict.get('equipment', None)),
                                               {}, #Equipment is not changeable either
                                               {}
                                               )
        self.equipment.obspy_equipment = self.equipment.to_obspy()
        
        das_channels = attributes_dict.get('channels', {})
        channel_default = das_channels.get('default', {})
        if channel_default:
            del das_channels['default']
        
        #v here is the attributes_dict of each das_channel
        self.channels = [Channel(k, v, locations, start_date, end_date, channel_default, channel_modifications)
                   for k, v in das_channels.items()]
        
    def __repr__(self):
        s = f'\n\nInstrumentation({len(self.channels)-1} channels)'
        return s


class Channel(object):
    """
    Channel Class.

    Corresponds to StationXML/obspy Channel, plus das_channel code
    """
    def __init__(self, channel_label, attributes_dict, locations, start_date, end_date, 
                 channel_template={}, channel_mods_dict={}):
        """
        :param instrument: instrument description
        :type instrument: ~class `Instrument`
        :param das_channel: the DAS channel the instrument is on
        :type das_channel: str
        :param orientation_code: orientation code of this channel/instrument
        :type orientation_code: str
        :param location_code: channel location code (if different from network)
        :type location_code: str, opt
        :param startdate: channel startdate (if different from network)
        :type startdate: str, opt
        :param enddate: channel enddate (if different from network)
        :type enddate: str, opt
        """
        if channel_label == "default":
            return
        
        if not attributes_dict:
            raise TypeError('No channel attributes')
        
        #Complete das channel fields with default fields. This will create a complete channel
        self.das_channel = ObsMetadata(self.complete_channel(attributes_dict, channel_template))
        
        #Location and orientation code should not be changed, so no channel modifications
        self.location_code = self.das_channel.get('location_code', "00") # default is always 00
        
        # This is just the one-letter string code, Although the attribute is redundant, it si needede before 
        # calling class Orientation init at the bottomo of this method
        self.orientation_code = self.get_orientation_code(attributes_dict.get('orientation_code', None))
   
        self.location = Location.get_location_from_code(locations, self.location_code, "channel",
                                                        self.channel_id_code)
                                                                             
        self.start_date = start_date
        self.end_date = end_date  
        
        selected_channel_mods = self.get_selected_channel_modifications(self.channel_id_code, channel_mods_dict)
        
        self.instrument = Instrument(self.das_channel, selected_channel_mods) 
        
        # this is the complete dictionary under key orientation code. If present, azimuth and dip can be changed in channel modifications
        orientation_dict = self.das_channel.get('orientation_code', None)
        if isinstance(orientation_dict, str):
            orientation_dict = {orientation_dict: {}} #If just a string, convert to dict to simplify code
            
        orientation_dict = ObsMetadata(orientation_dict)
        #Channel polarity is defined only while creating Instrument object. Don't change order of assignments      
        self.orientation = Orientation(orientation_dict, selected_channel_mods, self.instrument.polarity)
        
        self.comments = attributes_dict.get("comments", [])
        self.extras = [str(k) + ": " + str(v) for k, v in (attributes_dict.get('extras', {})).items()]
        self.convert_notes_and_extras_to_obspy()
        
        self.obspy_channel = self.to_obspy()
        
        
    @property
    def channel_id_code(self):
        """
        Identify channel uniquely through a combination of orientation and location code
        in that order, separated by a dash
        """
        
        if not self.orientation_code:
            raise TypeError(f'No orientation code in channel')

        # If no location, use 00 as defaul
        return self.orientation_code + "_" + (self.location_code if self.location_code else "00")
    
    def get_orientation_code(self, orientation_dict):
        """
        Get the orientation code from a dict that may include azimuth and dip
        """
        if isinstance(orientation_dict, str):
            keys = orientation_dict
        else:
            keys = list(orientation_dict.keys())
        #There should be only one key
        return keys[0]
        
    
    def get_selected_channel_modifications(self, id_code, channel_mods_dict):
        """
        Select one among several possible channel_modifications
        """

        # Get general default
        default_channel_mod = channel_mods_dict.get("*", {})
        if not default_channel_mod:
            default_channel_mod = channel_mods_dict.get("*-*", {})
        # Get defaults by location and orientation
        default_channel_loc = channel_mods_dict.get(self.orientation_code + "-*", {})
        default_channel_orient = channel_mods_dict.get("*-" + self.location_code, {})
        
        # Get modifications for this particular channel
        chmod = channel_mods_dict.get(id_code, {})
        if not chmod:
            # If id code not found, try with just the orientation part
            chmod = channel_mods_dict.get(id_code[0:1], {})
            
        # Gather all modifications in a single channel_mods
        # Do this in order: particular mods have priority over orientation-specific which has
        # priority over location-specific which has priority over general default
        for k, v in default_channel_loc.items():
            if k not in chmod:    
               chmod[k] = v
        for k, v in default_channel_orient.items():
            if k not in chmod:    
               chmod[k] = v    
        for k, v in default_channel_mod.items():
            if k not in chmod:    
               chmod[k] = v  
        
        return chmod      
           
    def complete_channel(self, das_channel, channel_default):
        """
        Take all the fields defined for each das channel and complement them with the channel_default fields
        If das_channel key exists, leave the value. If not, add channel_default key/value
        """
        #If there are no modifications, use default
        if not das_channel:
            return channel_default
        
        # First fill out allt he channel attributes 
        # Only get channel template if das_channel attribute does not exist
        for k, v in channel_default.items():
            if k not in das_channel:    
               das_channel[k] = v
        
        return das_channel
    
    def __repr__(self):
        s = f'\n\n\nChannel({self.channel_id_code}, '
        s += f'orientation code="{self.orientation.orientation_code}"'
        if self.location:
            s += f',location={self.location}'

        if self.start_date:
            s += f', startdate={self.start_date}'
 
        if self.end_date:
            s += f', enddate={self.end_date}'        
        s += ')'
        return s


    def channel_code(self, sample_rate):
        """
        Return channel code for a given sample rate

        :param sample_rate: instrumentation sampling rate (sps)i
        :kind sample_rate: float
        """
        inst_code = self.instrument.sensor.seed_instrument_code
        
        if len(inst_code) != 1:
            raise ValueError(f'Instrument code "{inst_code}" is not a single letter')
        if self.orientation.orientation_code not in ["X", "Y", "Z", "1", "2", "3", "H", "F"] :
            raise ValueError(f'Orientation code "{self.orientation_code}" is not a single letter')
        
        return (self._band_code_validation(sample_rate)
                + inst_code
                + self.orientation.orientation_code)

    def _band_code_validation(self, sample_rate):
        """
        Return the channel band code

        :param sample_rate: sample rate (sps)
        """
        bbc = self.instrument.sensor.seed_band_base_code
        if len(bbc) != 1:
            raise ValueError(f'Band base code "{bbc}" is not a single letter')
            
        if bbc in "FCHBMLVUWRPTQ":
            if sample_rate >= 1000:
                return "F"
            elif sample_rate >= 250:
                return "C"
            elif sample_rate >= 80:
                return "H"
            elif sample_rate >= 10:
                return "B"
            elif sample_rate > 1:
                return "M"
            elif sample_rate > 0.3:
                return "L"
            elif sample_rate >= 0.1:
                return "V"
            elif sample_rate >= 0.01:
                return "U"
            elif sample_rate >= 0.001:
                return "W"
            elif sample_rate >= 0.0001:
                return "R"
            elif sample_rate >= 0.00001:
                return "P"
            elif sample_rate >= 0.000001:
                return "T"
            else:
                return "Q"
        elif bbc in "GDES":
            if sample_rate >= 1000:
                return "G"
            elif sample_rate >= 250:
                return "D"
            elif sample_rate >= 80:
                return "E"
            elif sample_rate >= 10:
                return "S"
            else:
                warnings.warn("Short period sensor sample rate < 10 sps")
        else:
            raise TypeError(f'Unknown band base code: "{bbc}"')

    @property
    def seed_code(self):
         return self.channel_code()
     
    def to_obspy(self):
        """
         Create obspy object for channel
        """
        
        code = self.channel_code(self.instrument.sample_rate)   
        preamp = self.instrument.preamplifier.equipment.obspy_equipment if self.instrument.preamplifier else None
        comments = [Comment(s) for s in self.comments]
           
        channel = obspy_Channel(code, self.location_code, 
                        latitude=self.location.obspy_latitude, 
                        longitude=self.location.obspy_longitude,
                        elevation=self.location.elevation, 
                        depth=self.location.depth_m, 
                        azimuth=self.orientation.azimuth, 
                        dip=self.orientation.dip, 
                        types=None, 
                        external_references=None, 
                        sample_rate=self.instrument.sample_rate, 
                        sample_rate_ratio_number_samples=None, #Used for very low sample rates
                        sample_rate_ratio_number_seconds=None, 
                        storage_format=None, 
                        clock_drift_in_seconds_per_sample=None, 
                        calibration_units=None, calibration_units_description=None, 
                        sensor=self.instrument.sensor.equipment.obspy_equipment,
                        pre_amplifier=preamp, 
                        data_logger=self.instrument.datalogger.equipment.obspy_equipment, 
                        equipments=None, #This is at the instrumentation level, will be assigned there
                        response=self.instrument.obspy_response, 
                        description=self.channel_id_code, 
                        comments=comments,
                        start_date=self.start_date, #OJO: these will probably be deprecated
                        end_date=self.end_date,
                        restricted_status=None, 
                        alternate_code=None, 
                        historical_code=None, 
                        data_availability=None, 
                        identifiers=None, 
                        water_level=None, 
                        source_id=None)
         
        return channel
    
    def convert_notes_and_extras_to_obspy(self):
        """
        In StationXML comments are found at the channel level and up. Convert notes and extras to comments.
        """
        
        self.comments += self.comments + ['Extra attributes (for documentation purposes only):'] + self.extras           
        
class Instrument(Channel):
    """
    Instrument Class.
    """
    def __init__(self, attributes_dict, channel_modif={}):
        """
        :param datalogger: datalogger information
        :type datalogger: ~class `Datalogger`
        :param sensor: sensor information
        :type sensor: ~class `Sensor`
        :param preamplifier: preamplifier information
        :type preamplifier: ~class `Preamplifier`, opt
        """
        
        if not attributes_dict:
            raise TypeError('No instrument attributes')
        
        datalogger_config_selector = attributes_dict.get_configured_element(
                                                                      'datalogger_configuration', 
                                                                      channel_modif,
                                                                      {},
                                                                      None)
        sensor_config_selector = attributes_dict.get_configured_element(
                                                                  'sensor_configuration', 
                                                                  channel_modif,
                                                                  {},
                                                                  None)
        preamplifier_config_selector = attributes_dict.get_configured_element(
                                                                        'preamplifier_configuration', 
                                                                        channel_modif,
                                                                        {},
                                                                        None)
        
        self.datalogger = InstrumentComponent.dynamic_class_constructor('datalogger', 
                                              attributes_dict,
                                              channel_modif,
                                              None, #This is delay correction, calculated later
                                              datalogger_config_selector)
        delay_correction = self.datalogger.delay_correction

        #delay correction is a property of dataloggers, but needs to be used in sensors and preamplifiers
        self.sensor = InstrumentComponent.dynamic_class_constructor('sensor', 
                                              attributes_dict,
                                              channel_modif,
                                              delay_correction,
                                              sensor_config_selector)
        self.preamplifier = InstrumentComponent.dynamic_class_constructor('preamplifier', 
                                              attributes_dict,
                                              channel_modif,
                                              delay_correction,
                                              preamplifier_config_selector)
        
        # add the three component response stages
        self.add_and_renumber_all_response_stages()
        #Validate inputs and outputs of complete response sequence and correct delay    
        self.integrate_component_stages_into_single_list(delay_correction)
   
        self.obspy_response = self.to_obspy()
             
        #Finally, add sensitivity
        self.add_sensitivity(self.obspy_response)  
        
    def __repr__(self):
        return f'\nInstrument(Polarity="{self.polarity}", Output sample rate={self.total_output_sample_rate})'
    
    def integrate_component_stages_into_single_list(self, delay_correction):
        """
        1) Renumber stages sequentially 
        2) Verify/set units and sample rates
        3) Assure same frequency is used for consecutive PZ filters
        4) Calculate global polarity of the whole set of response stages
        5) Set global response delay correction
        6) Validate sample_rate expressed in datalogger component is equal to global response sample rate
        """
        
        # Stack response stages
        stages = self.response_stages
        
        last_stage = stages[0]
        last_frequency = last_stage.filter.normalization_frequency if last_stage.filter.type == 'PolesZeros' else None
        last_polarity = last_stage.polarity
        partial_accumulated_sample_rate = last_stage.output_sample_rate
        
        for this_stage in stages[1:]:
            
            """
            2a) Verify continuity of units
            """
            if last_stage.output_units != this_stage.input_units:
                raise ValueError(
                    f'Stage {stage.stage_sequence_number} and {this_stage.stage_sequence_number} units don\'t match')
            
            """
            2b) Verify continuity of sample rate. Calculate delay if necessary
            """
            if last_stage.input_sample_rate:
                if not this_stage.decimation_factor:
                    warnings.warn(f'No decimation factor for stage {this_stage.stage_sequence_number}, \
                                 assuming = 1 ')
                    this_stage.decimation_factor = 1
                    
                next_input_rate = last_stage.input_sample_rate / this_stage.decimation_factor
               
                if this_stage.input_sample_rate:
                    if this_stage.output_sample_rate != next_input_rate:
                        raise ValueError(
                            f'stage {this_stage.stage_sequence_number}  \
                            sample rate {this_stage.output_sample_rate} != expected {next_input_rate}')
                     
                else: # If no inpput sample rate specified (e.g. in FIR filters with decimation factor only), 
                      #store partial accumulated sample rate
                    this_stage.input_sample_rate = partial_accumulated_sample_rate
                    
                #Calculate delay, if absent, using real input sample rate
                this_stage.calculate_delay(delay_correction)
                    
            """
            3) Station XML requires that all PZ stages have the same normalization frequency.
            Check this condition
            """
            if last_frequency and last_frequency != 0 and this_stage.filter.type == 'PolesZeros':
                if last_frequency != this_stage.filter.normalization_frequency:
                    warnings.warn(
                        f'Normalization frequencies for PZ stages {stage.stage_sequence_number}\
                        and {this_stage.stage_sequence_number} don\'t match')
                last_frequency = this_stage.filter.normalization_frequency
                                 
            """
            4) Calculate global polarity
            """
            if not this_stage.polarity: #default polarity equals positive polarity
                this_stage.polarity = 1
                
            last_polarity *= this_stage.polarity
            
            last_stage = this_stage
            last_frequency = last_stage.filter.normalization_frequency if last_stage.filter.type == 'PolesZeros' else None
            partial_accumulated_sample_rate = last_stage.output_sample_rate     
            
        """
        5) Out of the loop: set global delay correction
        Apply global delay correction to the last stage (datalogger) if delay_correction = none
        All previous corrections in stages are zero in this case, so delay correcton is only applied once, at the end
        """
        stages[-1 ].correction = delay_correction \
            if delay_correction != None else last_stage.delay

        """
        6) Out of the loop: check global output sample rate
        """
        total_expected_sample_rate = partial_accumulated_sample_rate   

        if total_expected_sample_rate != self.sample_rate:
            warnings.warn(f'Datalogger declared sample rate {self.sample_rate} is different from calculated overall \
                sample rate of stages {total_expected_sample_rate}')
            
        """
        Set global response attributes
        """ 
        self.polarity = last_polarity 
        self.total_output_sample_rate = partial_accumulated_sample_rate
     
    def add_and_renumber_all_response_stages(self):
        
        """
        Adds all response stages as obsinfo and obpsy objects and renumbers the obsinfo ones
        """
 
        response_st = self.sensor.response_stages.stages
        obspy_st = self.sensor.response_stages.obspy_stages
        
        if self.preamplifier and self.preamplifier.response_stages:
            response_st += self.preamplifier.response_stages.stages
            obspy_st += self.preamplifier.response_stages.obspy_stages

        
        response_st += self.datalogger.response_stages.stages
        obspy_st += self.datalogger.response_stages.obspy_stages
        
        # Order the stage_sequence_numbers
        for i in range(len(response_st)):
            response_st[i].stage_sequence_number = i
           
        self.response_stages = response_st
        self.obspy_stages = obspy_st
                
    def to_obspy(self):
        """
        Return equivalent obspy response class
        """
        
        obspy_stages = self.obspy_stages
        
        obspy_response = Response(resource_id=None, 
                                  instrument_sensitivity=None, 
                                  instrument_polynomial=None, 
                                  response_stages=obspy_stages)
             
        return obspy_response
    
    
    def add_sensitivity(self, obspy_response):
        """
        Adds sensitivity to an obspy Response object

        Based on ..misc.obspy_routines.response_with_sensitivity
        """

        response_stg = self.response_stages
        true_input_units = input_units = response_stg[0].input_units
        output_units = response_stg[-1].output_units

        if "PA" in true_input_units.upper():
            # MAKE OBSPY THINK ITS M/S TO CORRECTLY CALCULATE SENSITIVITY
            input_units = "M/S"
        gain_prod = 1.
        for stage in response_stg:
            gain_prod *= stage.gain
              
        sensitivity = obspy_Sensitivity(
            gain_prod,
            response_stg[0].gain_frequency, #This could be provided according to StationXML but we assume it's equal to the gain frequency of first stage
            input_units=input_units,
            output_units=output_units, 
            input_units_description=response_stg[0].input_units_description,
            output_units_description=response_stg[-1].output_units_description
            )
        
        obspy_response.instrument_sensitivity = sensitivity
        #obspy_response.recalculate_overall_sensitivity(sensitivity.frequency)
        obspy_response.instrument_sensitivity.input_units = true_input_units
        obspy_response.instrument_sensitivity.output_units = output_units

       
    @property
    def equipment_datalogger(self):
        return self.datalogger.equipment
    @property
    def equipment_sensor(self):
        return self.sensor.equipment
    @property
    def equipment_preamplifier(self):
        return self.preamplifier.equipment
        
    def get_response_stage(self, num):    
    # All response stages are at the instrument_component level
        stages = self.response_stages
        assert(num <= stages[-1].stage_sequence_number), 'response stage out of range: {num}'
        
        return stages[num]
    
    @property
    def sample_rate(self):    
        return self.datalogger.sample_rate
    @property
    def delay_correction(self):  
        return self.delay_correction
    @property
    def seed_band_base_code(self):    
        return self.sensor.seed_band_base_code
    @property
    def seed_instrument_code(self): 
        return self.sensor.seed_instrument_code   
    @property
    def seed_orientation(self):    
        return self.orientation
    

class Location(object):
    """
    Location Class.
    """
    def __init__(self, attributes_dict):
        """
        Create Location object and assign attributes from attributes_dict.
        Validate required location attributes exist
        Convert to obspy longitude and latitude
        :param latitude: station latitude (degrees N)
        :type latitude: float
        :param longitude: station longitude (degrees E)
        :type longitude: float
        :param elevation: station elevation (meters above sea level)
        :type elevation: float
        :param uncertainties_m: uncertainties of [lat, lon, elev] in METERS
        :type uncertainties_m: list
        :param geology: site geology
        :type geology: str
        :param vault: vault type
        :type vault: str
        :param depth_m: depth of station beneath surface (meters)
        :type depth_m: float
        :param localisation_method: method used to determine position
        :type localisation_method: str
        """
        
        position = attributes_dict.get('position', None)
        base = attributes_dict.get('base', None)

        if not base:
            raise TypeError('No base in location')
        if not position:
            raise TypeError('No position in location')
                  
        self.latitude =  position.get('lat', None)
        self.longitude = position.get('lon', None)
        self.elevation = position.get('elev', None)
        self.uncertainties_m = base.get('uncertainties.m', None)
        self.geology = base.get('geology', None)
        self.vault = base.get('vault', None)
        self.depth_m = base.get('depth.m', None)
        self.localisation_method = base.get('localisation_method', None)
        self.obspy_latitude, self.obspy_longitude = self.get_obspy_latitude_and_longitude()
        
    def __repr__(self):
        discontinuous = False
        s = f'Location({self.latitude:g}, {self.longitude:g}, '
        s += f'{self.elevation:g}, {self.uncertainties_m}'
        if not self.geology == 'unknown':
            s += f', "{self.geology}"'
        else:
            discontinuous = True
        if self.vault:
            if discontinuous:
                s += f', vault="{self.vault}"'
            else:
                s += f', "{self.vault}"'
        else:
            discontinuous = True
        if self.depth_m:
            if discontinuous:
                s += f', depth_m={self.depth_m:g}'
            else:
                s += f', {self.depth_m}'
        else:
            discontinuous = True
        if self.localisation_method:
            if discontinuous:
                s += f', localisation_method="{self.localisation_method}"'
            else:
                s += f', "{self.localisation_method}"'
        else:
            discontinuous = True
        s += ')'
        return s
    
    @staticmethod
    def get_location_from_code(locations, code, type, label):
        """
         Obtain from the locations dictionary one location by code (key). Raise exception if not found
        """
        
        if not locations:
            raise TypeError(f'No locations specified in station {label}')
        
        loc = locations.get(code, None)

        if not loc:
            if type == 'station': #For station the location code is mandatory
                raise TypeError(f'Location "{code}" not found in station {label}')
            else:
                warnings.warn(f'Channel {label} has no location code. Assuming location code "00"')
                loc = locations.get("00", None)
                if not loc:
                    raise TypeError(f'Location "{code}" not found in channel {label}') 
                                
        return loc
        
    
    def get_obspy_latitude_and_longitude(self, location=None):
        """
        Returns obspy (latitude, longiutde) as a tupe.
        """
    
        if not location:
            location = self
            
        #Convert meters to kilometers and then to degrees.
        #This is along a meridian, so it's a great circle, Assumes 40000 km Earth diameter    
        uncertainty_lat = kilometers2degrees(
                           location.uncertainties_m.get('lat', 0) / 1000 if location.uncertainties_m else 0 )
        latitude = Latitude(value=location.latitude, 
                            lower_uncertainty=uncertainty_lat,
                            upper_uncertainty=uncertainty_lat,
                            datum=None
                            )
        
        intermediate_uncert_lon = location.uncertainties_m.get('lon', 0) / 1000 if location.uncertainties_m  else 0
        #lon is along parallel, not a great circle, so we must reduce by cosine.. Using numpy
        #Won't work for poles, make uncertainty zero in that case
        uncertainty_lon = kilometers2degrees(intermediate_uncert_lon * numpy.cos(latitude)) if latitude != 90 else 0
                           
        longitude = Longitude(value=self.longitude, 
                              lower_uncertainty=uncertainty_lon,
                              upper_uncertainty=uncertainty_lon,
                              datum=None
                              )
        return (latitude, longitude)

   
class Operator(Channel):
    """  
    Operator Class.
    """
   
    def __init__(self, attributes_dict):
        """
        :param attributes_dict: datalogger information
        :type attributes_dict: ~class `Datalogger`
        """
       
        if not attributes_dict:
            return None 
        
        self.reference_name = attributes_dict.get('reference_name', None)
        self.full_name = attributes_dict.get('full_name', None)
        contact_name = attributes_dict.get('contact', None)
        
        if contact_name:
           self.contact = contact_name.get('first_name', '') \
                          + " " + contact_name.get('last_name', '')
        else:
           self.contact = ""
        
        self.email = attributes_dict.get('email', None)
        self.country_code, self.area_code, self.phone_number = self.convert_phone_number(attributes_dict.get('phone_number', None))
        
        self.website = attributes_dict.get('website', None)
        
    def __repr__(self):
        s = f'\nOperator(Reference Name={self.reference_name}, Full Name = {self.full_name}, '
        if self.contact:
            s += f'Contact Name={self.contact}, '
        s += f'Email={self.email}, phone_number={self.phone_number}, website={self.website})'
        
        return s
    
    def convert_phone_number(self, phone):
        """
        Try to convert international nubmers to the horrible FDSN American standard, 
        """
        
        country_code = None
        area_code = "0" 
        phone_number = ""
    
        if not phone:
            return (country_code, area_code, phone_number)
         
        if isinstance(phone, dict):
            country_code = phone.get('country_code', None)
            area_code = phone.get('area_code', "")
            phone_number = phone.get('phone_number', "")
        elif isinstance(phone, type("")):
            #For reference:
            #country = re.compile("^(\+{0,1}|00)[0-9]{1,3}$")
            #area = re.compile("^\({0,1}[0-9]{3}\){0,1}$")
            #phone = re.compile("^[0-9]{3}\-[0-9]{4}$")
            
            us_phone_ptrn = '(?P<country>(\\+{0,1}|00)[0-9]{1,3}) *(?P<area>\\({0,1}[0-9]{3}\\){0,1}) *(?P<phone>[0-9]{3}\\-[0-9]{4})$'
            us_phone_re = re.compile(us_phone_ptrn)
             
            us_phone_match = us_phone_re.match(phone)
            if us_phone_match:
                country_code = us_phone_match.group['country']
                area_code = us_phone_match.group['area']
                phone_number = us_phone_match.group['phone']            
            else: 
                c_code_plus_ptn = "^\+([0-9]{1,3})"
                c_code_zero_ptn = "^00([0-9]{1,3})"
                c_code_plus_re= re.compile(c_code_plus_ptn)
                c_code_zero_re = re.compile(c_code_zero_ptn)
                phone_ptn = "(?P<phone>(\([0-9]+\))* *([0-9]+[ \-\.]*)*[0-9]+)$"
                
                c_code = c_code_plus_re.match(phone)
                if not c_code: #tried to do it with | for alternatives in reg exp, doesn't work
                    c_code = c_code_zero_re.match(phone)
                    phone_re = re.compile(c_code_zero_ptn + " *" + phone_ptn)
                else:
                    phone_re = re.compile(c_code_plus_ptn + " *" + phone_ptn) 
                 
                if c_code:
                    country_code = c_code.group(1)
                    
                mtch = phone_re.match(phone)
                if mtch:
                    phone_number = mtch.group('phone')
                    # The following is done to avoid FDSN reg exp restrictions for phones, American based
                    for chr in ["(", ")", ".", "-", " "]: 
                        phone_number = phone_number.replace(chr, "")
                    phone_number = phone_number[0:3] + "-" + phone_number[3:]
                    
        else:
            pass
                
        return (country_code, area_code, phone_number)  
        
        
class Orientation(object):
    """
    Class for sensor orientations. No channel modifs. Cannot change orientation as it is part of the channel identifiers. Azimuth and dip can be changed
    """
    def __init__(self, attributes_dict, channel_modifs, polarity):
        """
        Constructor
        Seismometer: 90º dip is down, positive voltage. Negative voltage  inverts to -90º (up)
        Hydrophone:  90º dip is down, positive pressure. Negative pressure  inverts to -90º (up)
        Seismometer vertical voltage positive = mouvement vers le haut
        Hydrophone voltage positive = augmentation de pression
        :param azimuth: azimuth (clockwise from north, degrees)
        :param azimuth_uncertainty: degrees
        :param dip: dip (degrees, -90 to 90: positive=down, negative=up)
        :param dip_uncertainty: degrees
        """
        if not attributes_dict:
            raise ValueError('No orientation')
        
        #if a dictionary attributes_dict contains azimuth and/of dip info else it's a simple string and is included in a list for generality
        keys = list(attributes_dict.keys() if isinstance(attributes_dict, dict) else attributes_dict) 
        if "1" in keys:
            self.orientation_code = "1" 
            value = ObsMetadata(attributes_dict.get("1", None))
            if not value:
                raise ValueError('Type "1" channel has no azimuth')  
            azimuth, azimuth_uncertainty = self.get_value_with_uncertainty(
                                               value.get_configured_element(
                                                   'azimuth.deg', 
                                                   channel_modifs,
                                                   {}, 
                                                   None))
            if azimuth == None:
                raise ValueError('Type "1" channel has no azimuth')
            dip, dip_uncertainty = [0, 0]
            
        elif "2" in keys:
            self.orientation_code = "2"  
            value = ObsMetadata(attributes_dict.get("2", None))
            if not value:
                raise ValueError('Type "2" channel has no azimuth')  
            azimuth, azimuth_uncertainty = self.get_value_with_uncertainty(
                                               value.get_configured_element(
                                                   'azimuth.deg', 
                                                   channel_modifs,
                                                   {}, 
                                                   None))
            if azimuth == None:
                raise ValueError('Type "2" channel has no azimuth')
            dip, dip_uncertainty = [0, 0] 
                                                            
        elif "3" in keys:
            self.orientation_code = "3"  
            value = ObsMetadata(attributes_dict.get("3", None))
            if not value:
                raise ValueError('Type "3" channel has no dip')  
            azimuth, azimuth_uncertainty = [0, 0]
            dip, dip_uncertainty = self.get_value_with_uncertainty(
                                        value.get_configured_element(
                                                 'dip.deg', 
                                                 channel_modifs,
                                                 {}, 
                                                 None))
            if dip == None:
                raise ValueError('Type "3" channel has no dip')
            
        elif "H" in keys:
            self.orientation_code = "H"  
            azimuth, azimuth_uncertainty = [0, 0]
            dip, dip_uncertainty = [90 * polarity, 0]
            
        elif "X" in keys:
            self.orientation_code = "X"  
            azimuth, azimuth_uncertainty = [0, 5] #Uncertainty defined by FDSN
            dip, dip_uncertainty = [0.0]
            
        elif "Y" in keys:
            self.orientation_code = "Y"  
            azimuth, azimuth_uncertainty = [90, 5] #Uncertainty defined by FDSN
            dip, dip_uncertainty = [0.0]
        
        elif "Z" in keys:
            self.orientation_code = "Z"  
            azimuth, azimuth_uncertainty = [0, 0] #Uncertainty defined by FDSN
            dip, dip_uncertainty = [90 * polarity, 5]
            
        else:
            raise ValueError('Type(s) "{keys}" orientation subcode(s) not implemented')
        
        self.azimuth = obspy_types.FloatWithUncertaintiesAndUnit(
            azimuth, lower_uncertainty=azimuth_uncertainty,
            upper_uncertainty=azimuth_uncertainty, unit='degrees')
        self.dip = obspy_types.FloatWithUncertaintiesAndUnit(
            dip, lower_uncertainty=dip_uncertainty,
            upper_uncertainty=dip_uncertainty, unit='degrees')
        
    
    def __repr__(self):
        s = f'\nOrientation('
        if self.azimuth:
            s += f', Azimuth={self.azimuth}'
        if self.dip:    
            s += f', Dip={self.dip}'
        
        return s

        
    def get_value_with_uncertainty(self, info_list):
        """
        Validate that info_list is a 2 member list, return the two members as a list, otherwise  [None, None]
        """
        
        if not info_list or not isinstance(info_list, list) or len(info_list) != 2:
            return [None, None]
        
        return [info_list[0], info_list[1]]
        


"""
Station class
"""
# Standard library modules
import warnings
warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
# Non-standard modules
from obspy.core.inventory.station import Station as obspy_Station

import obspy.core.util.obspy_types as obspy_types
from obspy.core.inventory.util import  (Latitude, Longitude, Site, Comment)
from obspy.core.utcdatetime import UTCDateTime
from obspy.taup.seismic_phase import self_tokenizer
# obsinfo modules
from ..instrumentation.instrumentation import (Instrumentation, Location)
from ..obsMetadata.obsmetadata import (ObsMetadata)

class Station(object):
    """
    Station. Equivalent to obspy/StationXML Station
    """
    def __init__(self, label, attributes_dict):
        """
        Constructor

        :param site: site description
        :kind site: str
        :param start_date: station start date
        :kind start_date: str
        :param end_date: station start date
        :kind end_date: str
        :param location_code: station location code (2 digits)
        :kind location_code: str
        :param locations: locations (names and positions)
        :kind locations: ~class `obsinfo.network.Locations`
        :param instruments: list of Instrumentation
        :kind instruments: list
        :param processing: processing steps
        :kind processing: dict (maybe should have class?)
        :param restricted_status: "open", "closed", "partial", or "unknown"
        :kind restricted_status: str
        """
         
        if not attributes_dict:
            raise TypeError('No station attributes')
        
        self.label = label
        
        self.site = attributes_dict.get("site", None)
        
        start_date = ObsMetadata.validate_date(attributes_dict.get("start_date", 0))
        self.start_date = UTCDateTime(start_date)
        
        end_date = ObsMetadata.validate_date(attributes_dict.get("end_date", 0))
        self.end_date = UTCDateTime(end_date)
        
        self.location_code = attributes_dict.get("location_code", None)
        
        self.restricted_status = attributes_dict.get("restricted_status", 'unknown')
        
        self.locations = {c: Location(v) for c, v in \
                          attributes_dict.get('locations', None).items()}
        self.location = Location.get_location_from_code(self.locations, self.location_code, "station", self.label)
      
        instr_list = attributes_dict.get('instrumentations', None)
        if instr_list and isinstance(instr_list, list):
            channel_mods = attributes_dict.get('channel_modifications', {})     
            self.instrumentation = [Instrumentation(inst, self.locations, start_date, end_date, channel_mods) 
                                    for inst in instr_list]
        else:
            instr_dict = attributes_dict.get('instrumentation', None)
            channel_mods = attributes_dict.get('channel_modifications', {})

            if instr_dict:
                self.instrumentation = Instrumentation(instr_dict, self.locations, start_date, end_date, channel_mods) 
            else:
                raise ValueError(f'No instrumentation in station {self.site}')
        #Locations, start_date and end_date are used to creat obspy Channel
              
        self.processing = attributes_dict.get("processing", None)
        
        self.comments = attributes_dict.get("commnents", [])
        self.extras = [str(k) + ": " + str(v) for k, v in (attributes_dict.get('extras', {})).items()]
        self.convert_notes_and_extras_to_obspy() 
        
        self.obspy_station = self.to_obspy()
        
        """
        Processing.from_attributes_dict(attributes_dict.get('processing',
                                                          None)),
        """

    def __repr__(self):
        s = f'\nStation(Label={self.label}, Site={self.site}, Start Date={self.start_date}, End Date={self.end_date}, '
        s += f'Location Code={self.location_code}, '
        s += f'{len(self.locations)} Locations, '
        if self.processing:
            s += f'{len(self.processing)} processing-steps'
        #if not self.restricted_stations == "unknown":
        #    s += f', {self.restricted_status}'
        s += ')'
        return s

    def to_obspy(self):
        """
         Convert station to obspy object
        """
        """
        Don't know what to do with 
         
        self.location.depth_m 
        self.location.localisation_method
        self.processing 
        """
            

        channels_number = len(self.instrumentation.channels)
        latitude, longitude = Location.get_obspy_latitude_and_longitude(self.location) 
        
        start_date = UTCDateTime(self.start_date if self.start_date else 0)
        end_date = UTCDateTime(self.end_date if self.end_date else 0)
        site = Site(name=self.site, description=None, town=None, county=None, region=None, country=None)
        comments = [Comment(s) for s in self.comments]
        
        obspy_station = obspy_Station( code = self.label, 
                                       latitude = latitude, 
                                       longitude = longitude, 
                                       elevation = self.location.elevation,
                                       channels=[ch.obspy_channel for ch in self.instrumentation.channels], 
                                       site=site, 
                                       vault=self.location.vault, 
                                       geology=self.location.geology, 
                                       equipments=[self.instrumentation.equipment], 
                                       operators=None,  #Will be assigned in class Network, where it is specified.
                                       creation_date=start_date, 
                                       termination_date=end_date, 
                                       total_number_of_channels=channels_number, 
                                       selected_number_of_channels=channels_number, 
                                       description=None, 
                                       comments=comments,
                                       start_date=start_date, 
                                       end_date= end_date, 
                                       restricted_status=self.restricted_status, 
                                       alternate_code=None, 
                                       historical_code=None, 
                                       data_availability=None, 
                                       identifiers=None, 
                                       water_level=None, 
                                       source_id=None)
            
        return obspy_station
    
    def convert_notes_and_extras_to_obspy(self):
        """
        Convert info file notes and extras to XML comments
        """
    
        self.comments += ['EXTRA ATTRIBUTES (for documentation only):'] + self.extras
    
    

class Processing(object):
    """
    Processing Class.

    Saves a list of Processing steps
    For now, just stores the list
    """
    def __init__(self, the_list):
        """
        :param the_list: list of processing steps
        :type list: list
        """
        self.list = the_list

    def __repr__(self):
        s = f'Processing({self.list})'
        return s
